import logging
from typing import Callable, NamedTuple, Optional

from ruteni import configuration
from ruteni.api import WebApi
from ruteni.plugins.database import database, metadata
from ruteni.plugins.locale import locales
from ruteni.plugins.session import get_session_user
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    Table,
    and_,
    event,
    func,
    text,
)
from sqlalchemy.engine.base import Connection
from sqlalchemy.sql import select
from sqlalchemy_utils import EmailType
from starlette.authentication import requires
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from urlobject.path import URLPath

logger = logging.getLogger(__name__)

ADMIN_USERNAME: str = configuration.get("RUTENI_ADMIN_USERNAME", default="admin")
ADMIN_EMAIL: str = configuration.get("RUTENI_ADMIN_EMAIL", default="admin")

users = Table(
    "users",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("display_name", String(32), nullable=False),
    Column("email", EmailType, nullable=False, unique=True),
    Column("locale_id", Integer, ForeignKey(locales.c.id), nullable=False),
    Column("added_at", DateTime, nullable=False, server_default=func.now()),
    Column("disabled_at", DateTime, default=None),
)

Index(
    "ix_users_email_not_disabled",
    users.c.email,
    unique=True,
    sqlite_where=users.c.disabled_at.is_(None),
    postgresql_where=users.c.disabled_at.is_(None),
)


def after_create(target: Table, connection: Connection, **kwargs):  # type: ignore
    connection.execute(
        text(
            "INSERT INTO %s (display_name,email,locale_id) VALUES ('%s','%s',1)"
            % (target.name, ADMIN_USERNAME, ADMIN_EMAIL)
        )
    )


event.listen(users, "after_create", after_create)


class UserInfo(NamedTuple):
    id: int
    display_name: str
    email: str
    locale: str

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            display_name=self.display_name,
            email=self.email,
            locale=self.locale,
        )


async def get_user_by_id(user_id: int) -> Optional[UserInfo]:
    row = await database.fetch_one(
        select([users, locales.c.code])
        .select_from(users.join(locales))
        .where(and_(users.c.id == user_id, users.c.disabled_at.is_(None)))
    )
    return (
        UserInfo(row["id"], row["display_name"], row["email"], row["code"])
        if row
        else None
    )


async def get_user_by_email(email: str) -> Optional[UserInfo]:
    row = await database.fetch_one(
        select([users, locales.c.code])
        .select_from(users.join(locales))
        .where(and_(users.c.email == email, users.c.disabled_at.is_(None)))
    )
    return (
        UserInfo(row["id"], row["display_name"], row["email"], row["code"])
        if row
        else None
    )


async def add_user(display_name: str, email: str, locale: str) -> UserInfo:
    locale_id = await database.fetch_val(
        select([locales.c.id]).where(locales.c.code == locale)
    )
    if locale_id is None:
        raise Exception(f"unknown locale {locale}")
    user_id = await database.execute(
        users.insert().values(
            display_name=display_name,
            email=email,
            locale_id=locale_id,
        )
    )
    return UserInfo(user_id, display_name, email, locale)


class UnknownUserException(Exception):
    def __init__(self, user_id: int) -> None:
        super().__init__(f"unknown user ID {user_id}")
        self.user_id = user_id


async def assert_user_exists(user_id: int) -> None:
    if await get_user_by_id(user_id) is None:
        raise UnknownUserException(user_id)


@requires("authenticated")
async def user_info(request: Request) -> Response:
    return JSONResponse(get_session_user(request))


UserTest = Callable[[int], bool]


class UserAccessMixin:
    "# TODO: abstract class?"

    def __init__(self, accessible_to: UserTest) -> None:
        self.accessible_to = accessible_to


api = WebApi("user", 1)
api.add_route(URLPath("info"), user_info, methods=["GET"], name="info")

logger.info("loaded")



from pathlib import Path
from typing import Optional

import PIL.IcoImagePlugin
import PIL.PngImagePlugin
from starlette.requests import Request
from starlette.responses import FileResponse, Response
from urlobject.path import URLPath

PURPOSE_VALUE_SET = set(("any", "monochrome", "maskable"))


class Icon:
    def __init__(self, ns: URLPath, path: URLPath, purpose: Optional[str] = None):
        self.ns = ns
        self.path = path
        if purpose and not set(purpose.split()).issubset(PURPOSE_VALUE_SET):
            raise ValueError("invalid purpose value")
        self.purpose = purpose  # TODO: store set?

    @property
    def src(self) -> str:
        return self.ns.add(self.path)

    @property
    def sizes(self) -> str:
        raise NotImplementedError

    @property
    def type(self) -> str:
        raise NotImplementedError

    @property
    def image(self) -> PIL.ImageFile.ImageFile:
        raise NotImplementedError

    def has(self, size: str) -> bool:
        return size in self.sizes.split()

    def to_dict(self) -> dict:
        result = dict(src=self.src, sizes=self.sizes, type=self.type)
        if self.purpose:
            result["purpose"] = self.purpose
        return result

    async def handle_request(self, request: Request) -> Response:
        raise NotImplementedError


class FileIcon(Icon):
    def __init__(
        self,
        ns: URLPath,
        path: URLPath,
        filename: Path,
        *,
        purpose: Optional[str] = None,
    ):
        super().__init__(ns, path, purpose=purpose)
        self.filename = filename

    async def handle_request(self, request: Request) -> Response:
        return FileResponse(self.filename, media_type=self.type)


class PngIcon(FileIcon):
    @property
    def sizes(self) -> str:
        return "{}x{}".format(*self.image.size)

    @property
    def type(self) -> str:
        return "image/png"

    @property
    def image(self) -> PIL.ImageFile.ImageFile:
        with open(self.filename, "rb") as fp:
            return PIL.PngImagePlugin.PngImageFile(fp=fp, filename=self.filename)


class IcoIcon(FileIcon):
    @property
    def sizes(self) -> str:
        return " ".join(
            f"{width}x{height}" for width, height in sorted(self.image.info["sizes"])
        )

    @property
    def type(self) -> str:
        return "image/x-icon"

    @property
    def image(self) -> PIL.ImageFile.ImageFile:
        with open(self.filename, "rb") as fp:
            return PIL.IcoImagePlugin.IcoImageFile(fp=fp, filename=self.filename)



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Any, Callable, List, Sequence, Set, Union

from pkg_resources import resource_filename
from starlette.applications import Starlette
from starlette.config import Config
from starlette.middleware import Middleware
from starlette.middleware.base import RequestResponseEndpoint
from starlette.responses import FileResponse
from starlette.routing import BaseRoute, Mount, Route
from starlette.staticfiles import PathLike, StaticFiles
from starlette.types import ASGIApp
from urlobject.path import URLPath

from .types import Callback, Service

# @todo https://nuculabs.dev/2021/05/18/fastapi-uvicorn-logging-in-production/
# logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
logger = logging.getLogger(__name__)

EndpointTransform = Callable[[RequestResponseEndpoint], RequestResponseEndpoint]

ROOTNS = URLPath("/")
STATICNS = ROOTNS.add("static")


class Configuration(Config):
    def __init__(self) -> None:
        super().__init__(os.environ.get("RUTENI_CONFIG", ".env"))
        self.routes: list[Union[Route, Mount]] = []
        self.middleware: list[Middleware] = []
        self.services: list[Service] = []
        self.static_dir = Path("/var/www/ruteni/static")

    def set(self, key: str, value: Any) -> None:
        self.environ[key] = value

    @property
    def env(self) -> str:
        return self.get("RUTENI_ENV", default="production")

    @property
    def is_devel(self) -> bool:
        return self.env == "development"

    @property
    def is_debug(self) -> bool:
        return self.get("RUTENI_DEBUG", cast=bool, default=False)

    def set_static_dir(self, static_dir: PathLike) -> Mount:
        # TODO: raise exception if in production
        self.static_dir = Path(static_dir)
        return self.add_static_mount("static", static_dir)

    def add_service(self, name: str, startup: Callback, shutdown: Callback) -> Service:
        service = Service(name=name, startup=startup, shutdown=shutdown)
        self.services.append(service)
        return service

    def add_middleware(self, cls: type, **options: Any) -> Middleware:
        middleware = Middleware(cls, **options)
        self.middleware.append(middleware)
        return middleware

    def add_route(
        self,
        path: URLPath,
        endpoint: RequestResponseEndpoint,
        *,
        ns: URLPath = ROOTNS,
        methods: List[str] = None,
        name: str = None,
        include_in_schema: bool = True,
    ) -> Route:
        route = Route(
            str(ns.add(path)),
            endpoint,
            methods=methods,
            name=name,
            include_in_schema=include_in_schema,
        )
        self.routes.append(route)
        return route

    def add_file_route(
        self,
        path: URLPath,
        filename: PathLike,
        *,
        ns: URLPath = ROOTNS,
        media_type: str = None,
        name: str = None,
    ) -> Route:
        endpoint: RequestResponseEndpoint = lambda request: FileResponse(
            filename, media_type=media_type
        )
        return self.add_route(path, endpoint, ns=ns, name=name, include_in_schema=False)

    def add_mount(
        self,
        path: URLPath,
        app: ASGIApp = None,
        routes: Sequence[BaseRoute] = None,
        name: str = None,
    ) -> Mount:
        mount = Mount(path, app, routes, name)
        self.routes.append(mount)
        return mount

    def add_static_mount(
        self,
        path: URLPath,
        directory: PathLike = None,
        packages: List[str] = None,
        html: bool = False,
        check_dir: bool = True,
        *,
        ns: URLPath = ROOTNS,
    ) -> Mount:
        return self.add_mount(
            ns.add(path),
            app=StaticFiles(
                directory=directory, packages=packages, html=html, check_dir=check_dir
            ),
        )

    def add_static_resource_mount(
        self,
        path: URLPath,
        name: str,
        packages: List[str] = None,
        html: bool = False,
        check_dir: bool = True,
    ) -> Mount:
        if not self.static_dir:
            directory = resource_filename(name, "www")
            return self.add_static_mount(path, directory, ns=STATICNS)

    def get_route_paths(self) -> Set[str]:
        return set(route.path for route in self.routes)


configuration = Configuration()


class Ruteni(Starlette):
    def __init__(self) -> None:
        super().__init__(
            debug=configuration.is_debug,
            on_startup=[self._start_services],
            on_shutdown=[self._stop_services],
            routes=configuration.routes,
            middleware=configuration.middleware,
        )
        self.shutdown_callbacks: list[Service] = []

    async def _start_services(self) -> None:
        try:
            for service in configuration.services:
                await service.startup(self)
                self.shutdown_callbacks.append(service)
        except BaseException:
            logger.exception("start")
            await self._stop_services()
            raise

    async def _stop_services(self) -> None:
        while len(self.shutdown_callbacks):
            service = self.shutdown_callbacks.pop()
            try:
                await service.shutdown(self)
            except Exception:
                logger.exception("stop")



from html.parser import HTMLParser
from typing import Any
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch

import socketio

from .config import (
    BASE_URL,
    DOMAIN,
    USER_EMAIL,
    USER_LOCALE,
    USER_NAME,
    USER_PASSWORD,
    clear_database,
    test_env,
)

with test_env:
    from ruteni import Ruteni
    import ruteni.plugins.registration
    from ruteni.plugins.verifications import VERIFICATION_MAX_ATTEMPTS


from .dns import DNSMock
from .server import UvicornTestServer
from .smtp import SMTPMock

NAMESPACE = "/ruteni/registration"
MX_SERVER = "mx." + DOMAIN


class VerificationEmailHTMLParser(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self.in_a = False
        self.code: str = ""

    def handle_starttag(self, tag: str, attrs: list) -> None:
        if self.code == "":
            self.in_a = tag == "a"

    def handle_endtag(self, tag: str) -> None:
        if self.code == "":
            self.in_a = False

    def handle_data(self, data: str) -> None:
        if self.code == "" and self.in_a and len(data) == 6:
            self.code = data


class RegistrationTestCase(IsolatedAsyncioTestCase):
    async def asyncSetUp(self) -> None:
        self.server = UvicornTestServer(Ruteni())
        await self.server.up()
        self.sio = socketio.AsyncClient(reconnection=False)
        await self.sio.connect(BASE_URL, transports=["websocket"], namespaces=NAMESPACE)

    async def asyncTearDown(self) -> None:
        await self.sio.disconnect()
        await self.sio.eio.http.close()  # TODO: socketio bug?
        await self.server.down()

    def tearDown(self) -> None:
        clear_database()

    async def call(self, *args, **kwargs) -> Any:  # type: ignore
        return await self.sio.call(*args, namespace=NAMESPACE, **kwargs)

    async def emit_edit(
        self, display_name: str, email: str, password: str, locale: str
    ) -> dict:
        return await self.call(
            "edit",
            {
                "display_name": display_name,
                "email": email,
                "password": password,
                "locale": locale,
            },
        )

    async def test_register(self) -> None:
        # send a bogus event
        self.assertEqual(await self.call("unknown"), {"error": "unknown-command"})

        # send bogus arguments
        self.assertEqual(await self.call("edit", None), {"error": "invalid-arguments"})
        self.assertEqual(await self.call("edit", 42), {"error": "invalid-arguments"})
        self.assertEqual(await self.call("edit", "a"), {"error": "invalid-arguments"})
        self.assertEqual(await self.call("edit", []), {"error": "invalid-arguments"})
        self.assertEqual(await self.call("edit", {}), {"error": "invalid-arguments"})
        self.assertEqual(
            await self.call("edit", {"email": ""}), {"error": "invalid-arguments"}
        )
        self.assertEqual(
            await self.call("edit", {"bogus-param": ""}), {"error": "invalid-arguments"}
        )

        self.assertEqual(
            await self.emit_edit("", "", "", ""),
            {
                "invalid-display-name": "empty",
                "invalid-email": "empty",
                "invalid-password": "empty",
                "invalid-locale": "empty",
            },
        )
        self.assertEqual(
            await self.emit_edit(100 * "x", 100 * "x", 100 * "x", ""),
            {
                "invalid-display-name": "overflow",
                "invalid-email": "overflow",
                "invalid-password": "overflow",
                "invalid-locale": "empty",
            },
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "incomplete-email", "", USER_LOCALE),
            {"invalid-email": "incomplete", "invalid-password": "empty"},
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "foo@", "", "en-US"),
            {"invalid-email": "incomplete", "invalid-password": "empty"},
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "fooé@", "", "de-DE"),
            {
                "invalid-email": "parse-error",
                "invalid-password": "empty",
                "invalid-locale": "unknown",
            },
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "@", "", USER_LOCALE),
            {"invalid-email": "parse-error", "invalid-password": "empty"},
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "@fr", "", USER_LOCALE),
            {"invalid-email": "parse-error", "invalid-password": "empty"},
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "foo@@", "bad", USER_LOCALE),
            {"invalid-email": "parse-error", "low-password-strength": 0},
        )
        self.assertEqual(
            await self.emit_edit(USER_NAME, "é@bar", "littlebetter", USER_LOCALE),
            {"invalid-email": "parse-error", "low-password-strength": 1},
        )

        dns = DNSMock({DOMAIN: [(MX_SERVER, 1, 1)]})
        with patch(
            "ruteni.plugins.registration.query_mx",
            wraps=dns.query_mx,
        ):
            self.assertEqual(
                await self.emit_edit(USER_NAME, "foo@é", "0xbar45", USER_LOCALE),
                {"invalid-email": "unknown-domain", "low-password-strength": 2},
            )

        with patch(
            "ruteni.plugins.registration.query_mx",
            wraps=dns.query_mx,
        ):
            self.assertEqual(
                await self.emit_edit(USER_NAME, USER_EMAIL, USER_PASSWORD, USER_LOCALE),
                {},
            )

        # now that the fields are okay, register using a mock SMTP function
        smtp = SMTPMock()
        with patch(
            "ruteni.plugins.verifications.send_mail",
            wraps=smtp.send_mail,
        ):
            self.assertTrue(await self.call("register"))

        # send a wrong verification code
        self.assertEqual(
            await self.call("verify", "xxxxxx"), VERIFICATION_MAX_ATTEMPTS - 1
        )
        self.assertEqual(
            await self.call("verify", "xxxxxx"), VERIFICATION_MAX_ATTEMPTS - 2
        )

        # get the email that was sent, extract the code and send it
        message = smtp.get_email(MX_SERVER)
        html = message.get_payload()
        parser = VerificationEmailHTMLParser()
        parser.feed(html)
        self.assertEqual(await self.call("verify", parser.code), None)

        # send a known event but in the wrong state
        self.assertEqual(await self.call("edit"), {"error": "unauthorized-command"})


if __name__ == "__main__":
    import unittest

    unittest.main()



from typing import Optional, Protocol

import aiodns


class Record(Protocol):
    host: str
    priority: int
    ttl: int
    type: str = "MX"


# from pycares import ares_query_mx_result

RecordList = list[Record]  # ares_query_mx_result]


async def query_mx(domain: str, timeout: int = 3) -> Optional[RecordList]:
    resolver = aiodns.DNSResolver(timeout=timeout)
    try:
        return await resolver.query(domain, "MX")
    except aiodns.error.DNSError:
        return None



import enum
import logging

import werkzeug
from ruteni import ROOTNS, configuration
from ruteni.api import WebApi
from ruteni.plugins.database import database, metadata
from ruteni.plugins.users import users
from sqlalchemy import Column, DateTime, Enum, ForeignKey, Integer, Table, func
from sqlalchemy_utils.types.json import JSONType
from starlette.requests import Request
from starlette.responses import FileResponse, JSONResponse, Response
from starlette.routing import Route
from starlette.staticfiles import PathLike
from urlobject.path import URLPath

logger = logging.getLogger(__name__)

# https://scotthelme.co.uk/a-new-security-header-expect-ct/
EXPECT_CT_MAX_AGE: int = configuration.get(
    "RUTENI_SECURITY_EXPECT_CT_MAX_AGE", cast=int, default=60  # 86400  # 1 day
)

# https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Strict-Transport-Security
# https://hstspreload.org/
HSTS_MAX_AGE: int = configuration.get(
    "RUTENI_SECURITY_HSTS_MAX_AGE", cast=int, default=63072000  # 2 years
)

CSP_REPORT_CONTENT_TYPE = "application/csp-report"
EXPECT_CT_CONTENT_TYPE = "application/expect-ct-report+json"
CSP_REPORT_KEY = "csp-report"
EXPECT_CT_KEY = "expect-ct-report"


class ReportType(enum.Enum):
    CSP = 1
    EXPECT_CT = 2


security_reports = Table(
    "security_reports",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("type", Enum(ReportType), nullable=False),
    Column("user_id", Integer, ForeignKey(users.c.id), nullable=True),
    Column("report", JSONType, nullable=False),
    Column("timestamp", DateTime, nullable=False, server_default=func.now()),
)


async def csp_report(request: Request) -> Response:
    """
    {
        "csp-report": {
            "document-uri": "http://localhost:8000/app/store/",
            "referrer": "",
            "violated-directive": "script-src-elem",
            "effective-directive": "script-src-elem",
            "original-policy": "default-src 'self'; worker-src 'self'; frame-src 'none'; child-src 'none'; object-src 'none'; require-trusted-types-for 'script'; report-uri /api/logging/v1/csp-report/;",
            "disposition": "enforce",
            "blocked-uri": "inline",
            "line-number": 16,
            "source-file": "http://localhost:8000/app/store/",
            "status-code": 200,
            "script-sample": "",
        }
    }
    """
    body = await request.json()
    report = body[CSP_REPORT_KEY]  # TODO: have a marshmallow schema to validate report?
    content_type = request.headers["content-type"]
    if content_type != CSP_REPORT_CONTENT_TYPE:
        logger.warning(f"invalid csp-report content type {content_type}")
    logger.debug(report, request.headers)
    user_id = request.user.id if request.user.is_authenticated else None
    await database.execute(
        security_reports.insert().values(
            type=ReportType.CSP, user_id=user_id, report=report
        )
    )
    return JSONResponse(True)


async def expect_ct(request: Request) -> Response:
    """
    https://shaunc.com/blog/article/implementing-a-reporturi-endpoint-for-expectct-and-other-headers~Xdf4cU8EurV1
    https://www.tpeczek.com/2017/05/preparing-for-chromes-certificate.html
    {
        "expect-ct-report": {
            "date-time": "2017-05-05T12:45:00Z",
            "hostname": "example.com",
            "port": 443,
            "effective-expiration-date": "2017-05-05T12:45:00Z",
            ...
        }
    }
    """
    body = await request.json()
    report = body[EXPECT_CT_KEY]
    content_type = request.headers["content-type"]
    if content_type != EXPECT_CT_CONTENT_TYPE:
        logger.warning(f"invalid expect-ct content type {content_type}")
    user_id = request.user.id if request.user.is_authenticated else None
    await database.execute(
        security_reports.insert().values(
            type=ReportType.EXPECT_CT, user_id=user_id, report=report
        )
    )
    return JSONResponse(True)


api = WebApi("security", 1)
api.add_route("csp-report", csp_report, methods=["POST"], name="csp-report")
api.add_route("expect-ct", expect_ct, methods=["POST"], name="expect-ct")

CSP_REPORT_URI = api.url_for("csp-report")
EXPECT_CT_URI = api.url_for("expect-ct")


class ContentSecurityPolicy(werkzeug.datastructures.ContentSecurityPolicy):
    """
    https://csp.withgoogle.com/docs/strict-csp.html
    """

    def __init__(
        self,
        connect: bool = False,
        img: bool = False,
        manifest: bool = False,
        script: bool = False,
        style: bool = False,
        worker: bool = False,
    ) -> None:
        super().__init__()
        self.report_uri = CSP_REPORT_URI
        self.base_uri = "'none'"
        self.default_src = "'none'"
        if connect:
            self.connect_src = "'self'"
        if img:
            self.img_src = "'self'"
        if manifest:
            self.manifest_src = "'self'"
        if script:
            self.script_src = "'self'"
        if style:
            self.style_src = "'self'"
        if worker:
            self.worker_src = "'self'"

    def to_header(self) -> str:
        return (
            super().to_header()
            + "; require-trusted-types-for 'script'; trusted-types default dompurify"
        )


def set_security_headers(
    response: Response,
    content_security_policy: ContentSecurityPolicy,
    *,
    csp_report_only: bool = False,
) -> Response:
    """
    https://web.dev/csp-xss/
    https://flask.palletsprojects.com/en/2.0.x/security/
    https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Strict-Transport-Security
    https://github.com/shieldfy/API-Security-Checklist/blob/master/README.md
    https://mechlab-engineering.de/2019/01/security-testing-and-deployment-of-an-api-release-your-flask-app-to-the-internet/
    """
    hsts = f"max-age={HSTS_MAX_AGE}; includeSubDomains; preload"
    ect = f'max-age={EXPECT_CT_MAX_AGE}, enforce, report-uri="{EXPECT_CT_URI}"'

    headers = response.headers
    headers[
        "Content-Security-Policy-Report-Only"
        if csp_report_only
        else "Content-Security-Policy"
    ] = content_security_policy.to_header()
    headers["Strict-Transport-Security"] = hsts
    headers["Expect-CT"] = ect
    headers["X-Content-Type-Options"] = "nosniff"
    headers["X-Frame-Options"] = "deny"
    headers["X-XSS-Protection"] = "1; mode=block"
    headers["Access-Control-Allow-Methods"] = "GET"
    return response


def add_html_file_route(
    path: URLPath,
    filename: PathLike,
    content_security_policy: ContentSecurityPolicy,
    *,
    ns: URLPath = ROOTNS,
    media_type: str = None,
    name: str = None,
) -> Route:
    return configuration.add_route(
        path,
        lambda request: set_security_headers(
            FileResponse(filename, media_type=media_type), content_security_policy
        ),
        ns=ns,
        name=name,
        include_in_schema=False,
    )


configuration.add_static_resource_mount("security", __name__)

# TODO: add middleware that checks that csp header is set on html?

logger.info("loaded")



import logging
from typing import Optional

from anyio import open_file
from jwcrypto.common import json_decode, json_encode
from jwcrypto.jwe import JWE
from jwcrypto.jwk import JWK
from jwcrypto.jwt import JWT, JWTExpired
from ruteni import configuration
from ruteni.utils.jwkset import JWKSet
from starlette.applications import Starlette

logger = logging.getLogger(__name__)

JWKEYS_FILE = configuration.get("RUTENI_JWKEYS_FILE")


class KeyManager:
    def __init__(self, key_file: str) -> None:
        self.key_file = key_file
        self.public_key: JWK = None
        self.private_key: JWK = None

    @staticmethod
    def is_private(keyset: JWKSet) -> bool:
        for key in keyset:
            return key.has_private  # only check first one
        return False

    def encrypt(self, payload: str) -> str:
        protected_header = {
            "alg": "RSA-OAEP-256",
            "enc": "A256CBC-HS512",
            "typ": "JWE",
            "kid": self.public_key.get("kid"),
        }
        jwetoken = JWE(
            payload.encode("utf-8"),
            recipient=self.public_key,
            protected=protected_header,
        )
        return jwetoken.serialize()

    def decrypt(self, token: str) -> Optional[dict]:
        # @todo we should clear the access_token cookie on error
        try:
            return json_decode(JWT(jwt=token, key=self.keyset).claims)
        except JWTExpired:
            return None

    def create_token(self, claims: dict) -> str:
        # header = dict(alg='RS256', typ='JWT', kid=key.key_id)
        kid = self.private_key.get("kid")
        header = '{"alg":"RS256","typ":"JWT","kid":"' + kid + '"}'
        jwt = JWT(header=header, claims=json_encode(claims))
        jwt.make_signed_token(self.private_key)
        return jwt.serialize()

    def get_claims(self, token: str, issuer: str) -> dict:
        claims = self.decrypt(token)
        assert claims
        if claims["iss"] != issuer:
            raise Exception(f"invalid issuer, expected {issuer}, got {claims['iss']}")
        return claims

    async def startup(self, starlette: Starlette) -> None:

        async with await open_file(self.key_file) as f:
            content = await f.read()

        self.keyset = JWKSet.from_json(content)

        if len(self.keyset) == 0:
            raise Exception("empty key file")

        keyset_ids = set(int(key.get("kid")) for key in self.keyset)

        # max(self.keyset, key=lambda item: int(item.get("kid")))

        kid: Optional[str] = None
        for key in self.keyset:
            if kid is None or int(kid) < int(key.get("kid")):
                kid = key.get("kid")

        if KeyManager.is_private(self.keyset):
            logger.debug(f"private key {kid} out of {keyset_ids} from {self.key_file}")
            self.private_key = key
            self.public_key = JWK()
            self.public_key.import_key(**json_decode(self.private_key.export_public()))
        else:
            logger.debug(f"installing public keys from {self.key_file}")
            self.public_key = key
        logger.info("started")

    async def shutdown(self, starlette: Starlette) -> None:
        logger.info("stopped")


jwkeys = KeyManager(JWKEYS_FILE)
configuration.add_service("jwkeys", jwkeys.startup, jwkeys.shutdown)

logger.info("loaded")



import json
import asyncio
from datetime import datetime, timezone
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch
from typing import Any

from ruteni import Ruteni
from starlette import status
from starlette.middleware.cors import ALL_METHODS
from starlette.testclient import TestClient

from .config import (
    BASE_URL,
    CLIENT_ID,
    SITE_NAME,
    USER,
    USER_EMAIL,
    USER_LOCALE,
    USER_NAME,
    USER_PASSWORD,
    clear_database,
    test_env,
)

SIGNIN_URL = BASE_URL + "/api/jauthn/v1/signin"
SIGNOUT_URL = BASE_URL + "/api/jauthn/v1/signout"

with test_env:
    from ruteni.plugins.jwkeys import jwkeys
    from ruteni.plugins.passwords import register_user
    from ruteni.plugins.token import (
        ACCESS_TOKEN_EXPIRATION,
        REFRESH_TOKEN_EXPIRATION,
        new_refresh_token,
    )
    from ruteni.plugins.auth.token import ACCESS_TOKEN_COOKIE_NAME


class TokenAuthTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.ruteni = Ruteni()

    def tearDown(self) -> None:
        clear_database()

    def valid_signin_info(self, content: dict[str, Any]) -> bool:
        return (
            "access_token" in content
            and "refresh_token" in content
            and "user" in content
        )

    def test_login(self) -> None:
        with TestClient(self.ruteni) as client:
            asyncio.run(
                register_user(USER_NAME, USER_EMAIL, USER_LOCALE, USER_PASSWORD)
            )
            response = client.get("/")

            # only POST allowed
            for method in ALL_METHODS:
                if method != "POST":
                    with client.request(method, SIGNIN_URL) as response:
                        self.assertEqual(
                            response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                        )

            with client.post(SIGNIN_URL, data=b"data") as response:
                self.assertEqual(
                    response.status_code, status.HTTP_415_UNSUPPORTED_MEDIA_TYPE
                )

            # json content type but bad with content
            with client.post(
                SIGNIN_URL,
                data=b"{",
                headers={"Content-Type": "application/json"},
            ) as response:
                self.assertEqual(
                    response.status_code, status.HTTP_422_UNPROCESSABLE_ENTITY
                )

            # multipart content type but with bad content

            with client.post(
                SIGNIN_URL,
                data=b"{",
                headers={"Content-Type": "multipart/form-data"},
            ) as response:
                self.assertEqual(
                    response.status_code, status.HTTP_422_UNPROCESSABLE_ENTITY
                )

            # send json without any field
            with client.post(SIGNIN_URL, json={}) as response:
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

            # multipart content type but with empty content
            for data in (b"", ""):
                for content_type in ("multipart/form-data", "application/json"):
                    with client.post(
                        SIGNIN_URL,
                        data=data,
                        headers={"Content-Type": content_type},
                    ) as response:
                        self.assertEqual(
                            response.status_code, status.HTTP_422_UNPROCESSABLE_ENTITY
                        )

            for data in (b"", ""):
                with client.post(
                    SIGNIN_URL,
                    data=data,
                    headers={"Content-Type": "application/x-www-form-urlencoded"},
                ) as response:
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

            # send json with email and password but a third parameter
            with client.post(
                SIGNIN_URL, json={"email": "", "password": "", "foo": ""}
            ) as response:
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

            # send form with email and password but a third parameter
            data = {"email": "", "password": "", "foo": ""}
            with client.post(SIGNIN_URL, data=data) as response:
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

            # send json with email and password but not all strings
            with client.post(
                SIGNIN_URL,
                json={"email": 1, "password": "", "client_id": 1},
            ) as response:
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

            # send form with all fields but wrongs ones
            data = {"email": "qux@foo.fr", "password": "baz", "client_id": CLIENT_ID}
            with client.post(SIGNIN_URL, data=data) as response:
                self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

            # send json with email and password but wrongs ones
            with client.post(
                SIGNIN_URL,
                json={
                    "email": "qux@foo.fr",
                    "password": "baz",
                    "client_id": CLIENT_ID,
                },
            ) as response:
                self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

            # send form with correct email but wrong password
            data = {
                "email": USER_EMAIL,
                "password": "bad-password",
                "client_id": CLIENT_ID,
            }
            with client.post(SIGNIN_URL, data=data) as response:
                self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

            # send json with correct email but wrong password
            with client.post(
                SIGNIN_URL,
                json={
                    "email": USER_EMAIL,
                    "password": "bad-password",
                    "client_id": CLIENT_ID,
                },
            ) as response:
                self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

            now = datetime.now(timezone.utc).astimezone()
            timestamp = int(now.timestamp())
            COMMOM_CLAIMS = {
                "sub": "2",
                "iat": timestamp,
                "exp": timestamp + ACCESS_TOKEN_EXPIRATION,
                "rte": timestamp + REFRESH_TOKEN_EXPIRATION,
                "orig_iat": 0,
                "iss": SITE_NAME,
                "email": USER_EMAIL,
                "display_name": USER_NAME,
                "scope": [],
                "client_id": CLIENT_ID,
            }

            # send form with correct email and password as form
            refresh_token = new_refresh_token()
            with patch("ruteni.plugins.token.now", wraps=lambda: now), patch(
                "ruteni.plugins.token.new_refresh_token",
                wraps=lambda: refresh_token,
            ):
                data = {
                    "email": USER_EMAIL,
                    "password": USER_PASSWORD,
                    "client_id": CLIENT_ID,
                }
                with client.post(SIGNIN_URL, data=data) as response:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    content = response.json()
                    self.assertEqual(
                        content,
                        {
                            "access_token": {"jti": "1", **COMMOM_CLAIMS},
                            "refresh_token": refresh_token,
                            "user": USER,
                        },
                    )
                    self.assertTrue(self.valid_signin_info(content))

            # send json with correct email and password as json
            # because refresh tokens must be unique, use a new one
            refresh_token = new_refresh_token()
            with patch("ruteni.plugins.token.now", wraps=lambda: now), patch(
                "ruteni.plugins.token.new_refresh_token",
                wraps=lambda: refresh_token,
            ):
                with client.post(
                    SIGNIN_URL,
                    json={
                        "email": USER_EMAIL,
                        "password": USER_PASSWORD,
                        "client_id": CLIENT_ID,
                    },
                ) as response:
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                    content = response.json()
                    self.assertEqual(
                        content,
                        {
                            "access_token": {"jti": "2", **COMMOM_CLAIMS},
                            "refresh_token": refresh_token,
                            "user": USER,
                        },
                    )
                    self.assertEqual(len(client.cookies.get_dict()), 1)
                    access_token = client.cookies[ACCESS_TOKEN_COOKIE_NAME]
                    self.assertEqual(
                        jwkeys.decrypt(access_token), {"jti": "2", **COMMOM_CLAIMS}
                    )

            # only POST allowed for signout
            for method in ALL_METHODS:
                if method != "POST":
                    with client.request(method, SIGNOUT_URL) as response:
                        self.assertEqual(
                            response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                        )

            with client.post(
                SIGNOUT_URL,
                json={"refresh_token": refresh_token},
            ) as response:
                self.assertEqual(response.status_code, status.HTTP_200_OK)


if __name__ == "__main__":
    import unittest

    unittest.main()



import unittest
from pathlib import Path

from ruteni import ROOTNS
from ruteni.utils.icon import IcoIcon, PngIcon
from urlobject.path import URLPath

# /usr/bin/identify tests/icons/icons.ico
ICO_FILENAME = Path("tests/icons/icons.ico")
PNG_FILENAME = Path("tests/icons/icon-512x512.png")
ICO_PATH = URLPath("ico")
PNG_PATH = URLPath("png")


class IconTestCase(unittest.TestCase):
    def test_ico(self) -> None:
        ico_icon = IcoIcon(ROOTNS, ICO_PATH, ICO_FILENAME)
        self.assertEqual(
            ico_icon.sizes,
            "72x72 96x96 120x120 128x128 144x144 180x180 192x192 256x256 512x512",
        )

    def test_png(self) -> None:
        png_icon = PngIcon(ROOTNS, PNG_PATH, PNG_FILENAME)
        self.assertEqual(png_icon.sizes, "512x512")
        self.assertEqual(png_icon.src, "/" + PNG_PATH)

    def test_purpose(self) -> None:
        for purpose in ("any", "maskable", "monochrome", "any maskable"):
            self.assertEqual(
                PngIcon(ROOTNS, PNG_PATH, PNG_FILENAME, purpose=purpose).purpose,
                purpose,
            )
        with self.assertRaises(ValueError):
            PngIcon(ROOTNS, PNG_PATH, PNG_FILENAME, purpose="foo")
        with self.assertRaises(ValueError):
            PngIcon(ROOTNS, PNG_PATH, PNG_FILENAME, purpose="any foo")



import logging

from databases import Database
from ruteni import configuration
from sqlalchemy import MetaData
from starlette.applications import Starlette

logger = logging.getLogger(__name__)

metadata = MetaData()

DATABASE_URL = configuration.get("RUTENI_DATABASE_URL")
database = Database(DATABASE_URL)


async def startup(starlette: Starlette) -> None:
    if configuration.is_devel:
        from sqlalchemy import create_engine

        engine = create_engine(DATABASE_URL)
        # if engine.dialect.name == "sqlite":
        metadata.create_all(engine)

    await database.connect()
    logger.info("started")


async def shutdown(starlette: Starlette) -> None:
    await database.disconnect()
    logger.info("stopped")


configuration.add_service("database", startup, shutdown)

logger.info("loaded")



import logging
from email.errors import MessageDefect, MessageParseError
from email.headerregistry import Address
from functools import reduce
from pathlib import Path
from typing import Any, Optional, Union

import limits.storage
import limits.strategies
import ruteni.plugins.logging
import ruteni.plugins.translation
from pkg_resources import resource_filename
from ruteni import configuration
from ruteni.plugins.database import database
from ruteni.plugins.locale import locales
from ruteni.plugins.passwords import register_user
from ruteni.plugins.security import ContentSecurityPolicy, add_html_file_route
from ruteni.plugins.session import get_user_from_environ
from ruteni.plugins.socketio import sio
from ruteni.plugins.users import get_user_by_email
from ruteni.plugins.verifications import EmailVerifier
from ruteni.utils.dns import RecordList, query_mx
from sqlalchemy.sql import select
from transitions.core import MachineError
from transitions.extensions.asyncio import AsyncMachine
from zxcvbn import zxcvbn

logger = logging.getLogger(__name__)

Issues = dict[str, Union[bool, int, str]]

MINIMUM_PASSWORD_STRENGTH: int = configuration.get(
    "RUTENI_REGISTRATION_MINIMUM_PASSWORD_STRENGTH", cast=int, default=3
)
MAXIMUM_DISPLAY_NAME_LENGTH: int = configuration.get(
    "RUTENI_REGISTRATION_MAXIMUM_DISPLAY_NAME_LENGTH", cast=int, default=40
)
MAXIMUM_EMAIL_LENGTH: int = configuration.get(
    "RUTENI_REGISTRATION_MAXIMUM_EMAIL_LENGTH", cast=int, default=40
)
MAXIMUM_PASSWORD_LENGTH: int = configuration.get(
    "RUTENI_REGISTRATION_MAXIMUM_PASSWORD_LENGTH", cast=int, default=40
)

KNOWN_LOCALES = ("en-EN", "fr-FR")  # TODO: from available emails

# https://limits.readthedocs.org
# https://haliphax.dev/2021/03/rate-limiting-with-flask-socketio/
# https://pypi.org/project/fastapi-limiter/

# uri = "redis+unix:///var/run/redis/redis-server.sock"
# options: dict = {}
# redis_storage = limits.storage.storage_from_string(uri, **options)
memory_storage = limits.storage.MemoryStorage()
one_per_second = limits.RateLimitItemPerSecond(1, 1)
moving_window = limits.strategies.MovingWindowRateLimiter(memory_storage)


class Registration:
    "User registration"

    def __init__(self) -> None:
        self.display_name: Optional[str] = None
        self.email: Optional[str] = None
        self.password: Optional[str] = None
        self.locale: Optional[str] = None
        self.records: Optional[RecordList] = None
        self.email_verifier: Optional[EmailVerifier] = None
        self.issues: dict[str, Issues] = dict(
            display_name={}, email={}, password={}, locale={}
        )

    async def input_ok(
        self, rv: list, *, display_name: str, email: str, password: str, locale: str
    ) -> bool:
        "Set the registration information"
        if self.display_name != display_name:
            issues = self.issues["display_name"]
            issues.clear()
            if display_name == "":
                issues["invalid-display-name"] = "empty"
            elif len(display_name) > MAXIMUM_DISPLAY_NAME_LENGTH:
                issues["invalid-display-name"] = "overflow"
            self.display_name = display_name

        if self.email != email:
            issues = self.issues["email"]
            issues.clear()
            if len(email) == 0:
                issues["invalid-email"] = "empty"
            elif len(email) > MAXIMUM_EMAIL_LENGTH:
                issues["invalid-email"] = "overflow"
            elif "@" not in email or (
                len(email) > 1 and email[-1] == "@" and "@" not in email[:-1]
            ):
                # input is something like `foo` or 'foo@'
                try:
                    email.encode("ascii")
                except UnicodeEncodeError:
                    issues["invalid-email"] = "parse-error"
                else:
                    issues["invalid-email"] = "incomplete"
            else:
                try:
                    address = Address(display_name=display_name, addr_spec=email)
                except (IndexError, MessageDefect, MessageParseError):
                    issues["invalid-email"] = "parse-error"
                else:
                    logger.debug(f"resolving MX for {address.domain}")
                    records = await query_mx(address.domain)
                    if records is None:
                        issues["invalid-email"] = "unknown-domain"
                    else:
                        logger.debug(f"found records: {records}")
                        if len(records):
                            self.records = sorted(records, key=lambda rec: rec.priority)
                        else:
                            issues["invalid-email"] = "misconfigured-domain"
            self.email = email

        if self.password != password:
            issues = self.issues["password"]
            issues.clear()
            if len(password) == 0:
                issues["invalid-password"] = "empty"
            elif len(password) > MAXIMUM_EMAIL_LENGTH:
                issues["invalid-password"] = "overflow"
            else:
                results = zxcvbn(password, user_inputs=[display_name, email])
                if results["score"] < MINIMUM_PASSWORD_STRENGTH:
                    issues["low-password-strength"] = results["score"]
            self.password = password

        if self.locale != locale:
            issues = self.issues["locale"]
            issues.clear()
            if locale == "":
                issues["invalid-locale"] = "empty"
            else:
                locale_id = await database.fetch_val(
                    select([locales.c.id]).where(locales.c.code == locale)
                )
                if locale_id is None:
                    issues["invalid-locale"] = "unknown"
            self.locale = locale

        issues = reduce(lambda a, b: {**a, **b}, self.issues.values())
        rv.append(issues)
        return len(issues) == 0

    async def sent_ok(self, rv: list) -> int:
        assert self.display_name and self.email and self.records

        # ignore registration if an active user with the same email exists
        user_info = await get_user_by_email(self.email)
        if user_info is not None:
            # TODO: do everything but send the email to protect users' privacy? by
            # timing a failed registration, one could deduce that a user has an account
            rv.append(False)
            return False

        template_path = Path(
            resource_filename(__name__, f"emails/{self.locale}/verification.html")
        )

        self.email_verifier = EmailVerifier()
        sent = await self.email_verifier.send_verification_code(
            self.display_name, self.email, self.records, template_path
        )
        rv.append(sent)
        return sent

    async def expired(self, rv: list, code: str) -> bool:
        assert self.email_verifier
        expired = self.email_verifier.code_expired()
        if expired:
            rv.append(-1)
        return expired

    async def too_many(self, rv: list, code: str) -> bool:
        assert self.email_verifier
        remaining_attempts = self.email_verifier.incr_attempts()
        rv.append(remaining_attempts)
        return remaining_attempts == 0

    async def code_ok(self, rv: list, code: str) -> bool:
        assert self.email_verifier
        correct = self.email_verifier.code_correct(code)
        logger.debug("code_ok?", correct)

        if correct:
            assert self.display_name and self.email and self.locale and self.password
            user_id, password_id = await register_user(
                self.display_name, self.email, self.locale, self.password
            )
            await self.email_verifier.user_verified(user_id)
            logger.debug(
                "account created:", user_id, self.display_name, self.email, password_id
            )

            # logger.debug(await group_manager.add_user_to_group(user_id, "admin"))

        # TODO: send error type (code incorrect or expired)
        if correct:
            rv.append(None)
        return correct


TRANSITIONS = (
    {"trigger": "edit", "source": "Bad", "dest": "Good", "conditions": "input_ok"},
    {"trigger": "edit", "source": "Bad", "dest": "Bad"},
    {"trigger": "edit", "source": "Good", "dest": "Good", "conditions": "input_ok"},
    {"trigger": "edit", "source": "Good", "dest": "Bad"},
    {"trigger": "register", "source": "Good", "dest": "Sent", "conditions": "sent_ok"},
    {"trigger": "register", "source": "Good", "dest": "Good"},
    {"trigger": "modify", "source": "Sent", "dest": "Good"},
    {"trigger": "retry", "source": "Sent", "dest": "Sent"},
    {"trigger": "verify", "source": "Sent", "dest": "Good", "conditions": "expired"},
    {"trigger": "verify", "source": "Sent", "dest": "Done", "conditions": "code_ok"},
    {"trigger": "verify", "source": "Sent", "dest": "Good", "conditions": "too_many"},
    {"trigger": "verify", "source": "Sent", "dest": "Sent"},
    {"trigger": "disconnect", "source": "*", "dest": "Disconnected"},
)
TRIGGERS = set(transition["trigger"] for transition in TRANSITIONS)
REGISTRATION_FSM = {
    "states": ("Bad", "Good", "Sent", "Done", "Disconnected"),
    "initial": "Bad",
    "transitions": TRANSITIONS,
}
NAMESPACE = "/ruteni/registration"


machines: dict[str, AsyncMachine] = {}


async def on_connect(sid: str, environ: dict[str, str]) -> bool:
    # if the user is connected, reject connection
    if get_user_from_environ(environ) is not None:
        return False

    registration = Registration()
    machines[sid] = AsyncMachine(registration, **REGISTRATION_FSM)
    logger.debug(f"{sid} connected")
    return True


async def on_disconnect(sid: str) -> None:
    await machines[sid].model.disconnect()
    del machines[sid]
    logger.debug(f"{sid} disconnected")


async def catch_all(event: str, sid: str, data: Optional[Any] = None) -> Any:
    logger.debug(event, machines[sid].model.state, data)
    if event not in TRIGGERS:
        return {"error": "unknown-command"}
    rv: list = []
    registration = machines[sid].model
    try:
        if data is None:
            await registration.trigger(event, rv)
        elif isinstance(data, dict):
            await registration.trigger(event, rv, **data)
        elif isinstance(data, list):
            await registration.trigger(event, rv, *data)
        else:
            await registration.trigger(event, rv, data)
    except MachineError:
        return {"error": "unauthorized-command"}
    except TypeError:
        # this exception will be raised if the command parameters are wrong, e.g.:
        #   input_ok() got an unexpected keyword argument 'invalid-display-name'
        # TODO: this exception is too broad
        return {"error": "invalid-arguments"}
    except Exception:
        logger.exception("unexpected")
        return {"error": "unexpected-error"}

    logger.debug("rv", rv)
    return tuple(rv)


sio.on("connect", on_connect, namespace=NAMESPACE)
sio.on("disconnect", on_disconnect, namespace=NAMESPACE)
sio.on("*", catch_all, namespace=NAMESPACE)

content_security_policy = ContentSecurityPolicy(
    script=True, connect=True, style=True, img=True
)

add_html_file_route(
    "ap/register",
    resource_filename(__name__, "/resources/index.html"),
    content_security_policy,
)


configuration.add_static_resource_mount("registration", __name__)

logger.info("loaded")



import logging

from pkg_resources import resource_filename
from ruteni import configuration
from urlobject.path import URLPath

from socketio import ASGIApp, AsyncRedisManager, AsyncServer

logger = logging.getLogger(__name__)


USE_REDIS: bool = configuration.get(
    "RUTENI_SOCKETIO_USE_REDIS", cast=bool, default=False
)
PATH: URLPath = configuration.get(
    "RUTENI_SOCKETIO_PATH", cast=URLPath, default="/socket.io"
)

if USE_REDIS:
    from ruteni.plugins.redis import REDIS_URL

    client_manager = AsyncRedisManager(REDIS_URL)
else:
    client_manager = None

sio = AsyncServer(async_mode="asgi", client_manager=client_manager)
app = ASGIApp(sio, socketio_path="")
configuration.add_mount(PATH, app)

configuration.add_static_resource_mount("socketio", __name__)

logger.info("loaded")



import logging
import quopri
from datetime import datetime, timedelta
from email import message_from_string
from email.utils import formataddr
from pathlib import Path
from random import randint

from ruteni import configuration
from ruteni.plugins.database import database, metadata
from ruteni.plugins.site import ABUSE_URL, SITE_NAME
from ruteni.plugins.users import users
from ruteni.utils.dns import RecordList
from ruteni.utils.jinja2 import get_template_from_path
from ruteni.utils.smtp import send_mail
from sqlalchemy import Column, DateTime, ForeignKey, Integer, Table, func

logger = logging.getLogger(__name__)

FROM_ADDRESS: str = configuration.get("RUTENI_VERIFICATION_FROM_ADDRESS")
VERIFICATION_CODE_LENGTH: int = configuration.get(
    "RUTENI_VERIFICATION_CODE_LENGTH", cast=int, default=6
)
VERIFICATION_CODE_EXPIRATION_MINUTES: int = configuration.get(
    "RUTENI_VERIFICATION_CODE_EXPIRATION_MINUTES", cast=int, default=1  # 20 minutes
)
VERIFICATION_MAX_ATTEMPTS: int = configuration.get(
    "RUTENI_VERIFICATION_MAX_ATTEMPTS", cast=int, default=3
)

verifications = Table(
    "verifications",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey(users.c.id), nullable=False),
    Column("sent_at", DateTime, nullable=False),
    Column("verified_at", DateTime, nullable=False, server_default=func.now()),
)


class EmailVerifier:
    async def send_verification_code(
        self,
        display_name: str,
        email: str,
        records: RecordList,
        template_path: Path,
    ) -> int:
        random_code = randint(0, 10 ** VERIFICATION_CODE_LENGTH - 1)
        self.verification_code = f"{{:0{VERIFICATION_CODE_LENGTH}}}".format(random_code)
        self.attempts = 0

        logger.debug(f"email template: {template_path}")
        params = {
            "email": email,
            "from": FROM_ADDRESS,
            "to": formataddr((display_name, email)),
            "display_name": display_name,
            "site_name": SITE_NAME,
            "verification_code": self.verification_code,
            "abuse_url": ABUSE_URL,
        }
        template = await get_template_from_path(template_path)
        content = await template.render_async(params)
        message = message_from_string(
            quopri.encodestring(content.encode("utf-8")).decode("utf-8")
        )
        logger.debug(message)
        # self.sent_at = datetime.now()
        # return VERIFICATION_CODE_EXPIRATION_MINUTES

        for record in records:
            success = await send_mail(message, record.host)
            if success:
                self.sent_at = datetime.now()
                return VERIFICATION_CODE_EXPIRATION_MINUTES

        # SMTPConnectError: 421 Your IP (82.66.70.213) is temporary blacklisted by an
        # anti-troyan rule, retry later and/or visit http://postmaster.free.fr/
        # https://forum.ubuntu-fr.org/viewtopic.php?id=2008675
        logger.warn("could not deliver email")
        return 0

    def incr_attempts(self) -> int:
        self.attempts += 1
        return VERIFICATION_MAX_ATTEMPTS - self.attempts

    def code_expired(self) -> bool:
        return datetime.now() - self.sent_at > timedelta(
            minutes=VERIFICATION_CODE_EXPIRATION_MINUTES
        )

    def code_correct(self, code: str) -> bool:
        return self.verification_code == code

    async def user_verified(self, user_id: int) -> int:
        return await database.execute(
            verifications.insert().values(user_id=user_id, sent_at=self.sent_at)
        )



import logging

from ruteni import configuration
from starlette.applications import Starlette
from paramiko import common, ServerInterface, Channel, PKey

logger = logging.getLogger(__name__)

# TODO: implement…
raise NotImplementedError("FIXME")


class Server(ServerInterface):
    def __init__(self) -> None:
        pass

    def check_channel_request(self, kind: str, chanid: int) -> int:
        # if kind == "session":
        return common.OPEN_SUCCEEDED

    def check_auth_publickey(self, username: str, key: PKey) -> int:
        return common.AUTH_SUCCESSFUL

    def get_allowed_auths(self, username: str) -> str:
        return "publickey"

    def check_channel_exec_request(self, channel: Channel, command: str) -> bool:
        logger.debug(command)
        return True


class SSHd:
    def __init__(self) -> None:
        pass

    async def startup(self, starlette: Starlette) -> None:
        logger.info("started")

    async def shutdown(self, starlette: Starlette) -> None:
        logger.info("stopped")


sshd = SSHd()
configuration.add_service("sshd", sshd.startup, sshd.shutdown)

logger.info("loaded")



from asyncio import Event, create_task
from typing import List, Optional

from ruteni import Ruteni
from uvicorn import Config, Server


class UvicornTestServer(Server):
    def __init__(self, ruteni: Ruteni, host: str = "127.0.0.1", port: int = 8000) -> None:
        self._startup_done = Event()
        super().__init__(config=Config(ruteni, host=host, port=port, log_level="error"))

    async def startup(self, sockets: Optional[List] = None) -> None:
        await super().startup(sockets=sockets)
        self.config.setup_event_loop()
        self._startup_done.set()

    async def up(self) -> None:
        self._serve_task = create_task(self.serve())
        await self._startup_done.wait()

    async def down(self) -> None:
        self.should_exit = True
        await self._serve_task



from importlib import import_module
from typing import Any


def load_symbol(dotpath: str) -> Any:
    """load symbol in module.  symbol is right-most segment"""
    module_, symbol = dotpath.rsplit(".", maxsplit=1)
    m = import_module(module_)
    return getattr(m, symbol)



import logging

from marshmallow import Schema, validate
from marshmallow.fields import Email, String
from ruteni.api import WebApi
from ruteni.plugins.auth import AuthPair, register_identity_provider
from ruteni.plugins.groups import get_user_groups
from ruteni.plugins.passwords import PASSWORD_MAX_LENGTH, check_password
from ruteni.plugins.session import UserType, del_session_user, set_session_user
from ruteni.plugins.token import MAX_CLIENT_ID_LENGTH
from ruteni.plugins.users import get_user_by_id
from ruteni.utils.form import get_form
from starlette import status
from starlette.authentication import AuthCredentials, BaseUser
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse, RedirectResponse, Response

logger = logging.getLogger(__name__)


# same code as in auth/token.py
class RuteniUser(BaseUser):
    def __init__(self, user: UserType) -> None:
        self.id = user["id"]
        self.email = user["email"]
        self.name = user["display_name"]
        self.locale = user["locale"]
        self.groups = user["groups"]

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.name


def authenticate(user: UserType) -> AuthPair:
    scopes = user["groups"] + ["authenticated"]
    return AuthCredentials(scopes), RuteniUser(user)


register_identity_provider("ruteni", authenticate)


class SignInSchema(Schema):
    email = Email(required=True)  # type: ignore
    password = String(required=True, validate=validate.Length(max=PASSWORD_MAX_LENGTH))
    client_id = String(
        required=True, validate=validate.Length(max=MAX_CLIENT_ID_LENGTH)
    )


async def signin(request: Request) -> Response:
    form = await get_form(request, SignInSchema)
    password_info = await check_password(form["email"], form["password"])

    if password_info is None:  # TODO: not clear what status code should be used
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    user_info = await get_user_by_id(password_info.user_id)
    assert user_info
    user = user_info.to_dict()
    user["provider"] = "ruteni"
    user["groups"] = await get_user_groups(user_info.id)

    response = JSONResponse(user)
    set_session_user(request, user)
    return response


async def signout(request: Request) -> Response:
    del_session_user(request)
    return RedirectResponse(url="/")


api = WebApi("auth", 1)
api.add_route("signin", signin, methods=["POST"], name="signin")
api.add_route("signout", signout, methods=["GET"], name="signout")

logger.info("loaded")



from ruteni.utils.smtp import Message


class SMTPMock:
    def __init__(self) -> None:
        self.allow_email = True
        self.messages: dict[str, list[Message]] = {}

    def allow_email_delivery(self, allow: bool) -> None:
        self.allow_email = allow

    async def send_mail(self, message: Message, host: str) -> bool:
        if self.allow_email:
            if host not in self.messages:
                self.messages[host] = []
            self.messages[host].append(message)
            return True
        else:
            return False

    def get_email(self, host: str) -> Message:
        return self.messages[host].pop(0)



from pathlib import Path
from typing import Any, Optional

from anyio import open_file

import jinja2


async def get_template_from_path(path: Path) -> jinja2.Template:
    async with await open_file(path) as f:
        content = await f.read()
    return jinja2.Template(content, enable_async=True)


def get_template_env(directory: Path) -> jinja2.Environment:
    loader = jinja2.FileSystemLoader(directory)
    return jinja2.Environment(loader=loader, autoescape=True, enable_async=True)



from __future__ import annotations

from pathlib import Path
from typing import Optional

from starlette.responses import FileResponse
from starlette.routing import Route
from urlobject.path import URLPath

from . import EndpointTransform, ROOTNS
from .api import WebApi
from .static import WebStatic
from .urlns import URLNamespace

APPNS = ROOTNS.add("app")


class WebApp(URLNamespace):
    """
    Manage a web app
    """

    url_format = "/app/{name}/"
    all_apps: list[WebApp] = []

    def __init__(
        self,
        name: str,
        version: int,
        *,
        transform: Optional[EndpointTransform] = None,
    ) -> None:
        ns = URLPath(WebApp.url_format.format(name=name, version=version))
        super().__init__(
            "app", name, version, ns, transform=transform, include_in_schema=False
        )
        self.api = WebApi(name, version, transform=transform)
        self.static = WebStatic(name, version, transform=transform)
        WebApp.all_apps.append(self)

    def add_static_file_route(self, path: URLPath, filename: Path) -> Route:
        return self.static.add_route(path, lambda request: FileResponse(filename))



import logging
from starlette.applications import Starlette
from apscheduler.schedulers.asyncio import AsyncIOScheduler

from ruteni import configuration

logger = logging.getLogger(__name__)

SCHEDULER_WAIT = configuration.get("RUTENI_SCHEDULER_WAIT", default=False)

scheduler = AsyncIOScheduler()


async def startup(starlette: Starlette) -> None:
    scheduler.start()
    logger.info("started")


async def shutdown(starlette: Starlette) -> None:
    # TODO: modules that import `scheduler` should not be able to shut it down
    scheduler.shutdown(wait=SCHEDULER_WAIT)
    logger.info("stopped")


configuration.add_service("scheduler", startup, shutdown)

logger.info("loaded")



import logging

import babel
import httpx
from authlib.oidc.discovery.well_known import get_well_known_url
from ruteni.api import WebApi
from ruteni.plugins.auth import AuthPair, register_identity_provider
from ruteni.plugins.groups import get_user_groups
from ruteni.plugins.oauth import oauth
from ruteni.plugins.session import UserType, del_session_user, set_session_user
from ruteni.plugins.users import add_user, get_user_by_email
from starlette.authentication import AuthCredentials, BaseUser
from starlette.requests import Request
from starlette.responses import PlainTextResponse, RedirectResponse, Response

logger = logging.getLogger(__name__)


class GoogleUser(BaseUser):
    """
    As of 2021/07/20, `user` is something like this:
    "iss": "https://accounts.google.com",
    "azp": "1084386621732-qjmkaojnut33[...]7q22me2p6qb1.apps.googleusercontent.com",
    "aud": "1084386621732-qjmkaojnut33[...]7q22me2p6qb1.apps.googleusercontent.com",
    "sub": "106637688186237713656",
    "hd": "undomaine.fr",
    "email": "jean.dupont@undomaine.fr",
    "email_verified": True,
    "at_hash": "ndyKkyw5_LQTsA9JKLim4A",
    "nonce": "uYnEP5r9BQK9ilJ4ruC7",
    "name": "Jean Dupont",
    "picture": "https://lh3.googleusercontent.com/a/AATXAJymU07Yz4XziVRLP3RQK3...",
    "given_name": "Jean",
    "family_name": "Dupont",
    "locale": "fr",
    "iat": 1625999571,
    "exp": 1626003171,
    """

    def __init__(self, user: dict) -> None:
        self.id = user["id"]
        self.email = user["email"]
        self.name = user["name"]
        self.given_name = user["given_name"]
        self.family_name = user["family_name"]
        self.picture = user["picture"]
        self.groups = user["groups"]
        self.locale = user["locale"]  # TODO: convert locale to 5-letter format

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.name


def authenticate(user: UserType) -> AuthPair:
    assert user["iss"] == "https://accounts.google.com"
    scopes = user["groups"] + ["authenticated"]
    return AuthCredentials(scopes), GoogleUser(user)


register_identity_provider("google", authenticate)


google = oauth.register(
    name="google",
    server_metadata_url=get_well_known_url(
        "https://accounts.google.com", external=True
    ),
    client_kwargs={"scope": "openid email profile"},
)


async def signin(request: Request) -> Response:
    redirect_uri = request.url_for("google-auth")
    try:
        return await google.authorize_redirect(request, redirect_uri)
    except httpx.ConnectTimeout:
        return PlainTextResponse(
            "Connection timeout redirecting to google. Please retry later."
        )


async def auth(request: Request) -> Response:

    # get the token from the URL and parse it to get the user data
    token = await google.authorize_access_token(request)
    user = await google.parse_id_token(request, token)

    # look for the user ID in the database from his/her email address
    user_info = await get_user_by_email(user["email"])

    # if the user is unknown, create a new profile
    if user_info is None:
        locale = user["locale"]
        if locale in babel.core.LOCALE_ALIASES:
            locale = babel.core.LOCALE_ALIASES[locale].replace("_", "-")
        user_info = await add_user(user["name"], user["email"], locale)

    # record the user ID and the groups the user belongs to
    user["id"] = user_info.id
    user["groups"] = await get_user_groups(user_info.id)
    user["provider"] = "google"

    logger.debug(f'{user["email"]} logged in; groups: {user["groups"]}')

    # record in the user data in the session
    set_session_user(request, dict(user))
    return RedirectResponse(url="/")


async def signout(request: Request) -> Response:
    del_session_user(request)
    return RedirectResponse(url="/")


api = WebApi("google", 1)
api.add_route("/signin", signin, name="signin")
api.add_route("/signout", signout, name="signout")
api.add_route("/auth", auth, name="auth")

logger.info("loaded")



from __future__ import annotations

from typing import Optional

from urlobject.path import URLPath

from . import EndpointTransform
from .urlns import URLNamespace


class WebStatic(URLNamespace):
    """
    Manage static resources
    """

    # url_format = "/static/{name}/v{version}"
    url_format = "/static/{name}"

    def __init__(
        self,
        name: str,
        version: int,
        *,
        transform: Optional[EndpointTransform] = None,
    ) -> None:
        ns = URLPath(WebStatic.url_format.format(name=name, version=version))
        super().__init__(
            "static", name, version, ns, transform=transform, include_in_schema=False
        )



import enum
import json
import logging
from pathlib import Path
from typing import Any, Mapping, Optional, Sequence, Union

from marshmallow import Schema, fields
from marshmallow.exceptions import ValidationError
from ruteni import EndpointTransform, configuration
from ruteni.app import WebApp
from ruteni.utils.color import Color
from ruteni.utils.icon import Icon
from ruteni.utils.locale import Locale, get_locale_from_request
from starlette.requests import Request
from starlette.responses import FileResponse, JSONResponse, Response
from urlobject.path import URLPath

logger = logging.getLogger(__name__)

# https://github.com/madskristensen/WebEssentials.AspNetCore.ServiceWorker/blob/master/src/Webmanifest/WebManifest.cs

WEBMANIFEST_MIME_TYPE = "application/manifest+json"
IMPORTMAP_MIME_TYPE = "application/importmap+json"


class IconSchema(Schema):
    src = fields.String(required=True)  # /static/store/images/icons/icon-144x144.png
    sizes = fields.String(required=True)  # 144x144
    type = fields.String(required=True)  # image/png
    # purpose = Enum(..., required=False)  monochrome maskable any


class InvalidManifest(Exception):
    pass


class ManifestSchema(Schema):
    """
    https://developer.mozilla.org/en-US/docs/Web/Manifest
    https://developers.google.com/web/fundamentals/web-app-manifest
    https://w3c.github.io/manifest/
    """

    background_color = fields.String(required=False)
    # categories = List(fields.String, required=False)
    description = fields.String(required=False)
    # dir = Enum(..., required=False)
    display = fields.String(required=False)
    # display_override
    # iarc_rating_id = fields.String(required=False)
    icons = fields.List(fields.Nested(IconSchema), required=True)
    lang = fields.String(required=False)
    name = fields.String(required=True)
    # orientation = Enum(..., required=False)
    # prefer_related_applications = Boolean(required=False)
    # protocol_handlers = Nested(..., required=False)
    # related_applications = Nested(required=False)
    scope = fields.String(required=False)
    # screenshots = Nested(..., required=False)
    # shortcuts = Nested(..., required=False)
    short_name = fields.String(required=False)
    start_url = fields.String(required=False)
    theme_color = fields.String(required=False)
    # splash_pages


def validate_manifest(manifest: Mapping) -> Mapping:
    schema = ManifestSchema()
    try:
        data = schema.load(manifest)
    except ValidationError:
        raise InvalidManifest()  # TODO: add parameters

    # TODO: the icon property must include a 192px and a 512px sized icons
    return data


def load_manifest(manifest_path: str) -> Mapping[str, Any]:
    with open(manifest_path) as f:  # TODO: async?
        return validate_manifest(json.load(f))


class RelatedApplication:
    def __init__(self, platform: str, id: str) -> None:
        self.platform = platform
        self.id = id

    def to_dict(self) -> dict:
        return dict(platform=self.platform, id=self.id)


class Display(enum.Enum):
    DISPLAY = "fullscreen"
    STANDALONE = "standalone"
    MINIMAL_UI = "minimal-ui"
    BROWSER = "browser"


class I18n:
    def __init__(
        self,
        *,
        full_name: Optional[str] = None,
        short_name: Optional[str] = None,
        description: Optional[str] = None,
        categories: Optional[Sequence[str]] = None,
    ):
        if not full_name and not short_name:
            raise ValueError("either full_name or short_name must be not empty")
        self.full_name = full_name
        self.short_name = short_name
        self.description = description
        self.categories = categories

    @property
    def name(self) -> str:
        if self.full_name is not None:
            return self.full_name
        else:
            assert self.short_name
            return self.short_name

    def to_dict(self) -> dict:
        result: dict[str, Union[str, Sequence[str]]] = dict()
        if self.full_name:
            result["name"] = self.full_name
        if self.short_name:
            result["short_name"] = self.short_name
        if self.description:
            result["description"] = self.description
        if self.categories:
            result["categories"] = self.categories
        return result


class ProgressiveWebApp(WebApp):
    """
    Manage a progressive web app
    https://web.dev/add-manifest/
    """

    def __init__(
        self,
        name: str,
        version: int,
        *,
        transform: Optional[EndpointTransform] = None,
        manifest_name: Optional[str] = None,
        resources_name: Optional[str] = None,
        sw_name: Optional[str] = None,
        scope: Optional[URLPath] = None,
        display: Optional[Display] = None,
        theme_color: Optional[Color] = None,
        background_color: Optional[Color] = None,
        prefer_related_applications: Optional[bool] = False,
    ) -> None:
        super().__init__(name, version, transform=transform)
        self.i18ns: dict[Locale, I18n] = {}
        self.icons: list[Icon] = []
        self.related_applications: list[RelatedApplication] = []
        self.manifest_name = manifest_name or f"{self.name}.webmanifest"
        self.resources_name = resources_name or "resources.json"
        self.sw_name = sw_name or "sw.js"
        self.scope = scope
        self.display = display
        self.theme_color = theme_color
        self.background_color = background_color
        self.prefer_related_applications = prefer_related_applications
        self.add_route(self.manifest_name, self.handle_manifest)

    @property
    def available_locales(self) -> Sequence[Locale]:
        return list(self.i18ns.keys())

    async def handle_manifest(self, request: Request) -> Response:
        locale = get_locale_from_request(request, self.available_locales)
        return JSONResponse(self.get_manifest(locale), media_type=WEBMANIFEST_MIME_TYPE)

    def set_service_worker(self, sw_path: Path) -> None:
        self.add_route(self.sw_name, lambda request: FileResponse(sw_path))
        self.add_route(  # TODO: fix
            self.sw_name + ".map", lambda request: FileResponse(str(sw_path) + ".map")
        )

    def set_resources(self, resources_path: Path) -> None:
        self.add_route(
            self.resources_name, lambda request: FileResponse(resources_path)
        )

    def add_icon(self, icon: Icon) -> None:
        self.icons.append(icon)
        self.static.add_route(icon.src, icon.handle_request)

    def add_i18n(
        self,
        locale: Locale,
        *,
        full_name: Optional[str] = None,
        short_name: Optional[str] = None,
        description: Optional[str] = None,
        categories: Optional[Sequence[str]] = None,
    ) -> None:
        self.i18ns[locale] = I18n(
            full_name=full_name,
            short_name=short_name,
            description=description,
            categories=categories,
        )

    def add_related_application(self, related_application: RelatedApplication) -> None:
        self.related_applications.append(related_application)

    def get_manifest(self, locale: Locale) -> dict:
        if locale not in self.i18ns:
            raise KeyError(f"unknown locale {locale}")

        # not sure where people got that some sizes are required
        # https://w3c.github.io/manifest/#icons-member
        # https://web.dev/installable-manifest/
        # https://stackoverflow.com/questions/48839338/
        if not any(icon.has("192x192") or icon.has("512x512") for icon in self.icons):
            raise ValueError("invalid icon list")

        result = dict(
            lang=str(locale),
            icons=[icon.to_dict() for icon in self.icons],
            start_url=str(self.ns),
            **self.i18ns[locale].to_dict(),
        )

        if self.display:
            result["display"] = self.display.value
        if self.theme_color:
            result["theme_color"] = self.theme_color
        if self.background_color:
            result["background_color"] = self.background_color
        if self.prefer_related_applications:
            result["prefer_related_applications"] = self.prefer_related_applications
        if self.related_applications:
            result["related_applications"] = [
                related_application.to_dict()
                for related_application in self.related_applications
            ]
        return result


configuration.add_static_resource_mount("pwa", __name__)

logger.info("loaded")



import email.message
import logging
from typing import Union

from aiosmtplib import SMTP
from aiosmtplib.errors import (
    SMTPConnectError,
    SMTPConnectTimeoutError,
    SMTPResponseException,
    SMTPServerDisconnected,
    SMTPTimeoutError,
)

logger = logging.getLogger(__name__)

Message = Union[email.message.EmailMessage, email.message.Message]


async def send_mail(
    message: Message, hostname: str, port: int = 25, timeout: int = 10
) -> bool:
    smtp = SMTP(hostname=hostname, port=port, timeout=timeout)
    try:
        logger.debug(f"connecting to {hostname}")
        await smtp.connect()
        logger.debug("connected; sending message")
        try:
            await smtp.send_message(message)
            logger.debug("message sent")
        finally:
            smtp.close()
            logger.debug("connection closed")
    except (
        SMTPConnectError,
        SMTPConnectTimeoutError,
        SMTPResponseException,
        SMTPServerDisconnected,
        SMTPTimeoutError,
    ) as exc:
        logger.error(exc)
        return False
    else:
        return True



import os
import tempfile
from test.support import EnvironmentVarGuard

from ruteni import configuration
from ruteni.utils.jwkset import KeyCollection

TEST_DIR = tempfile.mkdtemp()
DATABASE_PATH = TEST_DIR + "/test.db"

BASE_URL = "http://localhost:8000"
CLIENT_ID = "client-id"
DOMAIN = "bar.fr"
USER_EMAIL = "username@" + DOMAIN
USER_NAME = "username"
USER_PASSWORD = "lWhgjgjgjHJy765"
USER_LOCALE = "fr-FR"
SITE_NAME = "Ruteni Test"

USER = {
    "id": 2,  # admin is 1
    "display_name": USER_NAME,
    "email": USER_EMAIL,
    "provider": "ruteni",
    "locale": "fr-FR",
    "groups": [],
}

SESSION_SECRET_KEY = "secret-key"
PRIVATE_KEYS = TEST_DIR + "/private_keys.json"
PUBLIC_KEYS = TEST_DIR + "/public_keys.json"
key_collection = KeyCollection(TEST_DIR, True)
key_collection.generate()
key_collection.export(PUBLIC_KEYS, PRIVATE_KEYS)

configuration.set("RUTENI_ENV", "development")
configuration.set("RUTENI_JWKEYS_FILE", PRIVATE_KEYS)
configuration.set("RUTENI_SITE_NAME", SITE_NAME)

test_env = EnvironmentVarGuard()
test_env.set("RUTENI_DATABASE_URL", "sqlite:///" + DATABASE_PATH)
configuration.set("RUTENI_VERIFICATION_FROM_ADDRESS", "Bar Baz <bar@baz.fr>")
configuration.set("RUTENI_VERIFICATION_ABUSE_URL", "<abuse_url>")
test_env.set("RUTENI_SESSION_SECRET_KEY", SESSION_SECRET_KEY)


def clear_database() -> None:
    os.unlink(DATABASE_PATH)



import aioredis
from ruteni import configuration

REDIS_URL: str = configuration.get(
    "RUTENI_REDIS_URL", default="redis+unix:///run/redis/redis-server.sock"
)

redis = aioredis.from_url(REDIS_URL)



import logging
from datetime import datetime, timezone

from marshmallow import Schema, validate
from marshmallow.fields import Email, String
from ruteni import configuration
from ruteni.api import WebApi
from ruteni.plugins.auth import AuthPair, register_identity_provider
from ruteni.plugins.cookie import set_cookie
from ruteni.plugins.groups import get_user_groups
from ruteni.plugins.passwords import PASSWORD_MAX_LENGTH, check_password
from ruteni.plugins.session import UserType
from ruteni.plugins.site import SITE_NAME
from ruteni.plugins.token import (
    MAX_CLIENT_ID_LENGTH,
    REFRESH_TOKEN_LENGTH,
    create_access_token,
    create_refresh_token,
    get_user_id_from_access_token,
    revoke_refresh_token,
    select_refresh_token,
)
from ruteni.utils.form import get_form
from starlette import status
from starlette.authentication import AuthCredentials, BaseUser
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

logger = logging.getLogger(__name__)

JWT_ISSUER: str = configuration.get("RUTENI_AUTH_JWT_ISSUER", default=SITE_NAME)
ACCESS_TOKEN_COOKIE_NAME: str = configuration.get(
    "RUTENI_AUTH_ACCESS_TOKEN_COOKIE_NAME", default="access_token"
)


# same code as in auth/session.py
class RuteniUser(BaseUser):
    def __init__(self, user: UserType) -> None:
        self.id = user["id"]
        self.email = user["email"]
        self.name = user["display_name"]
        self.locale = user["locale"]
        self.groups = user["groups"]

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.name


def authenticate(user: UserType) -> AuthPair:
    scopes = user["groups"] + ["authenticated"]
    return AuthCredentials(scopes), RuteniUser(user)


register_identity_provider("ruteni", authenticate)


class SignInSchema(Schema):
    email = Email(required=True)  # type: ignore
    password = String(required=True, validate=validate.Length(max=PASSWORD_MAX_LENGTH))
    client_id = String(
        required=True, validate=validate.Length(max=MAX_CLIENT_ID_LENGTH)
    )


async def signin(request: Request) -> Response:
    form = await get_form(request, SignInSchema)
    password_info = await check_password(form["email"], form["password"])

    if password_info is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    user_info, access_token, refresh_token = await create_refresh_token(
        password_info.user_id, JWT_ISSUER, form["client_id"]
    )

    user = user_info.to_dict()
    user["provider"] = "ruteni"
    user["groups"] = await get_user_groups(user_info.id)

    result = {
        "access_token": access_token.claims,
        "refresh_token": refresh_token,
        "user": user,
    }
    logger.debug(f"signin response: {result}")

    response = JSONResponse(result)
    set_cookie(response, ACCESS_TOKEN_COOKIE_NAME, access_token.token)
    return response


class RefreshTokenSchema(Schema):
    refresh_token = String(
        required=True, validate=validate.Length(equal=2 * REFRESH_TOKEN_LENGTH)
    )


async def signout(request: Request) -> Response:
    form = await get_form(request, RefreshTokenSchema)

    access_token = request.cookies.get(ACCESS_TOKEN_COOKIE_NAME)
    if not access_token:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    # TODO: if the access token is expired, this will fail
    user_id = get_user_id_from_access_token(access_token, JWT_ISSUER)
    success = await revoke_refresh_token(user_id, form["refresh_token"])

    response = JSONResponse(success)
    response.delete_cookie(ACCESS_TOKEN_COOKIE_NAME)
    return response


async def refresh(request: Request) -> Response:
    form = await get_form(request, RefreshTokenSchema)

    # TODO: implement…
    raise NotImplementedError("FIXME")

    row = await select_refresh_token(form["refresh_token"])

    if row is None:
        # raise ServiceUserException("invalid-token")
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    if row["disabled_reason"] is not None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        # raise ServiceUserException(
        #     "account-disabled",
        #     public=dict(reason=row["disabled_reason"]),
        #     private=dict(id=row["id"], info=disabled),
        # )

    if row["revoked_at"] is not None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        # raise ServiceUserException(
        #     "revoked-token",
        #     public=dict(at=row["revoked_at"]),
        #     private=dict(id=row["id"]),
        # )

    if row["expires_at"] < datetime.now(timezone.utc).astimezone():
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)
        # raise ServiceUserException(
        #     "expired-token",
        #     public=dict(at=expires_at),
        #     private=dict(id=row["id"]),
        # )

    access_token = await create_access_token(
        row["id"],
        row["expires_at"],
        row["user_id"],
        JWT_ISSUER,
        row["client_id"],
    )

    response = JSONResponse({"access_token": access_token.claims})
    set_cookie(response, ACCESS_TOKEN_COOKIE_NAME, access_token.token)
    return response


api = WebApi("jauthn", 1)
api.add_route("signin", signin, methods=["POST"])
api.add_route("signout", signout, methods=["POST"])
api.add_route("refresh", refresh, methods=["POST"])

logger.info("loaded")



import asyncio
from unittest import IsolatedAsyncioTestCase

from ruteni import Ruteni
from starlette import status
from starlette.middleware.cors import ALL_METHODS
from starlette.testclient import TestClient

from .config import (
    CLIENT_ID,
    USER,
    USER_EMAIL,
    USER_LOCALE,
    USER_NAME,
    USER_PASSWORD,
    clear_database,
    test_env,
)

with test_env:
    from ruteni.plugins.passwords import register_user
    from ruteni.plugins.session import get_user_from_cookie
    from ruteni.plugins.auth.session import api as session_api
    from ruteni.plugins.users import api as user_api

SIGNIN_URL = session_api.url_for("signin")
SIGNOUT_URL = session_api.url_for("signout")
USER_INFO_URL = user_api.url_for("info")


class SessionAuthTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.ruteni = Ruteni()

    def tearDown(self) -> None:
        clear_database()

    def test_login(self) -> None:
        with TestClient(self.ruteni) as client:
            asyncio.run(
                register_user(USER_NAME, USER_EMAIL, USER_LOCALE, USER_PASSWORD)
            )

            # signin with wrong methods
            for method in ALL_METHODS:
                if method != "POST":
                    response = client.request(method, SIGNIN_URL)
                    self.assertEqual(
                        response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                    )

            # signin with bad parameters
            response = client.post(
                SIGNIN_URL, json={"email": "", "password": "", "client_id": ""}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

            # signin with an unknown email
            response = client.post(
                SIGNIN_URL,
                json={"email": "foo@bar.fr", "password": "", "client_id": CLIENT_ID},
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

            # signin with a bad password
            response = client.post(
                SIGNIN_URL,
                json={
                    "email": USER_EMAIL,
                    "password": "bad-password",
                    "client_id": CLIENT_ID,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

            # signin with correct credentials
            response = client.post(
                SIGNIN_URL,
                json={
                    "email": USER_EMAIL,
                    "password": USER_PASSWORD,
                    "client_id": CLIENT_ID,
                },
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.json(), USER)
            self.assertEqual(get_user_from_cookie(client.cookies["session"]), USER)

            # get user info when authenticated
            response = client.get(USER_INFO_URL)
            self.assertEqual(response.json(), USER)

            # signout
            for method in ALL_METHODS:
                if method not in ("GET", "HEAD"):
                    response = client.request(method, SIGNOUT_URL)
                    self.assertEqual(
                        response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                    )

            response = client.get(SIGNOUT_URL)
            self.assertEqual(client.cookies.get_dict(), {})

            # get user info when not authenticated
            response = client.get(USER_INFO_URL)
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


if __name__ == "__main__":
    import unittest

    unittest.main()



import typing

from starlette.applications import Starlette

Callback = typing.Callable[[Starlette], typing.Awaitable[None]]


class Service(typing.NamedTuple):
    name: str
    startup: Callback
    shutdown: Callback



import logging
from typing import NamedTuple, Optional

from argon2 import PasswordHasher
from argon2.exceptions import VerifyMismatchError
from ruteni import configuration
from ruteni.plugins.database import database, metadata
from ruteni.plugins.users import add_user, users
from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    Table,
    and_,
    func,
)
from sqlalchemy.sql import select

logger = logging.getLogger(__name__)

PASSWORD_MAX_LENGTH: int = configuration.get(
    "RUTENI_PASSWORD_MAX_LENGTH", cast=int, default=40
)

passwords = Table(
    "passwords",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey(users.c.id), nullable=False),
    Column("hashed_password", String, nullable=False),  # TODO: length as pref?
    Column("added_at", DateTime, nullable=False, server_default=func.now()),
    Column("disabled_at", DateTime, default=None),
)


Index(
    "ix_passwords_user_id_not_disabled",
    passwords.c.user_id,
    unique=True,
    sqlite_where=passwords.c.disabled_at.is_(None),
    postgresql_where=passwords.c.disabled_at.is_(None),
)


ph = PasswordHasher()


class Credential(NamedTuple):
    user_id: int
    hashed_password: str


async def get_credential(email: str) -> Optional[Credential]:
    row = await database.fetch_one(
        select([users.c.id, passwords.c.hashed_password])
        .select_from(users.join(passwords))
        .where(
            and_(
                users.c.email == email,
                users.c.disabled_at.is_(None),
                passwords.c.disabled_at.is_(None),
            )
        )
    )
    return Credential(row["id"], row["hashed_password"]) if row else None


async def check_password(email: str, password: str) -> Optional[Credential]:
    credential = await get_credential(email)
    if credential is None:
        return None
    logger.debug("signin:", email, password, "~", credential.hashed_password)
    try:
        if ph.verify(credential.hashed_password, password):
            return credential
    except VerifyMismatchError:
        pass
    return None


async def add_password(user_id: int, password: str) -> int:
    return await database.execute(
        passwords.insert().values(user_id=user_id, hashed_password=ph.hash(password))
    )


async def register_user(
    display_name: str, email: str, locale: str, password: str
) -> tuple[int, int]:
    user_info = await add_user(display_name, email, locale)
    password_id = await add_password(user_info.id, password)
    return user_info.id, password_id


logger.info("loaded")



import unittest

from ruteni.utils.color import Color
from webcolors import CSS3_NAMES_TO_HEX


class ColorTestCase(unittest.TestCase):
    def test_color(self) -> None:
        with self.assertRaises(ValueError):
            Color("#f")
        with self.assertRaises(ValueError):
            Color("foo")
        for name, hex_value in CSS3_NAMES_TO_HEX.items():
            self.assertEqual(Color(name), name)
            self.assertEqual(Color(hex_value), hex_value)



import json

from marshmallow.exceptions import ValidationError
from multipart.multipart import parse_options_header
from starlette import status
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response
from typing import Type
from marshmallow import Schema


async def get_body(request: Request) -> Response:
    content_type_header = request.headers.get("Content-Type")
    content_type, options = parse_options_header(content_type_header)
    if content_type == b"application/json":
        try:
            return await request.json()
        except json.decoder.JSONDecodeError:
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY)
    elif (
        content_type == b"multipart/form-data"
        or content_type == b"application/x-www-form-urlencoded"
    ):
        try:
            return await request.form()
        except TypeError:
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY)
    else:
        raise HTTPException(status_code=status.HTTP_415_UNSUPPORTED_MEDIA_TYPE)


async def get_form(request: Request, Schema: Type[Schema]) -> dict:
    body = await get_body(request)
    schema = Schema()
    try:
        return schema.load(body)
    except ValidationError:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST)



import json
import logging
import os
from pathlib import Path

import ruteni.plugins.auth
import ruteni.plugins.quotquot
from pkg_resources import resource_filename
from ruteni import STATICNS, configuration
from ruteni.app import WebApp
from ruteni.components import load_component_dir
from ruteni.plugins.pwa import Display, ProgressiveWebApp
from ruteni.plugins.security import ContentSecurityPolicy, set_security_headers
from ruteni.plugins.users import UserAccessMixin
from ruteni.utils.color import Color
from ruteni.utils.icon import PngIcon
from ruteni.utils.jinja2 import get_template_env
from ruteni.utils.locale import Locale, get_html_lang, get_locale_from_request
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse, Response

logger = logging.getLogger(__name__)

# https://medium.com/@applification/progressive-web-app-splash-screens-80340b45d210

MANIFEST_NAME = "store.webmanifest"

app_store = ProgressiveWebApp(
    "store",
    1,
    manifest_name=MANIFEST_NAME,
    theme_color=Color("#2196f3"),
    background_color=Color("#2196f3"),
    display=Display.STANDALONE,
)


def get_resource(path: str) -> Path:
    return Path(resource_filename(__name__, "resources/" + path))


app_store.set_service_worker(get_resource("sw.js"))
app_store.set_resources(get_resource("resources.json"))

for size in (192, 512):
    name = f"images/icons/icon-{size}x{size}.png"
    icon = PngIcon(
        ns=app_store.static.ns,
        path=name,
        filename=configuration.static_dir / "store" / name,
        purpose="any maskable",
    )
    app_store.add_icon(icon)

app_store.add_i18n(
    Locale("en", "US"),
    full_name="Ruteni app store",
    short_name="app store",
    description="A simple app store",
    categories=["app", "store"],
)

app_store.add_i18n(
    Locale("fr", "FR"),
    full_name="Magasin d'applications de Ruteni",
    short_name="Magasin d'applications",
    description="Un magasin d'applications simple",
    categories=["applications", "magasin"],
)

template_dir = resource_filename(__name__, "templates")
template_env = get_template_env(Path(template_dir))

content_security_policy = ContentSecurityPolicy(
    connect=True, img=True, script=True, manifest=True
)


async def homepage(request: Request) -> Response:
    locale = get_locale_from_request(request, app_store.available_locales)
    i18n = app_store.i18ns[locale]
    params = dict(
        title=i18n.name,
        lang=get_html_lang(locale),
        manifest=MANIFEST_NAME,
        description=i18n.description,
        theme_color=app_store.theme_color,
        apple_touch_icon=app_store.icons[0].src,  # TODO: FIXME
    )
    homepage_template = template_env.get_template("index.j2")
    content = await homepage_template.render_async(params)
    return set_security_headers(HTMLResponse(content), content_security_policy)


app_store.add_route("", homepage)

configuration.add_static_resource_mount("store", __name__)


# list
async def list_apps(request: Request) -> Response:
    result: list = []
    for app in WebApp.all_apps:
        # ignore the app store itself
        if app is app_store:
            continue

        # if the app requires special access rights, check that the user statifies them
        if isinstance(app, UserAccessMixin) and not (
            request.user.is_authenticated and app.accessible_to(request.user.id)
        ):
            continue

        if isinstance(app, ProgressiveWebApp):
            locale = get_locale_from_request(request, app.available_locales)
            app_info = app.get_manifest(locale)
        else:
            app_info = dict(name=app.name)

        result.append(app_info)
    return JSONResponse(result)


app_store.api.add_route("list", list_apps)

logger.info("loaded")



from typing import NamedTuple, Optional

from ruteni.utils.dns import RecordList


class Record(NamedTuple):
    host: str
    priority: int
    ttl: int
    type: str = "MX"


class DNSMock:
    def __init__(self, domains: dict[str, list[tuple[str, int, int]]]) -> None:
        self.domains = domains

    async def query_mx(self, domain: str) -> Optional[RecordList]:
        if domain not in self.domains:
            return None
        return [Record(*fields) for fields in self.domains[domain]]  # type: ignore



from pkg_resources import resource_filename
from ruteni import configuration
from ruteni.static import WebStatic
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from starlette.responses import FileResponse, Response


ABUSE_URL: str = configuration.get("RUTENI_VERIFICATION_ABUSE_URL")
SITE_NAME: str = configuration.get("RUTENI_SITE_NAME")
LOGO_PATH: str = configuration.get(
    "RUTENI_SITE_LOGO_PATH", default=resource_filename(__name__, "/resources/logo.svg")
)
FAVICON_PATH: str = configuration.get(
    "RUTENI_SITE_FAVICON_PATH",
    default=resource_filename(__name__, "/resources/favicon.ico"),
)

static = WebStatic("site", 1)
static.add_file_route("logo.svg", LOGO_PATH)


# TODO: use an Icon from ruteni.utils.icon


class Favicon(HTTPEndpoint):
    # favicon = b64decode(
    #     "AAABAAEAEBACAAEAAQBWAAAAFgAAAIlQTkcNChoKAAAADUlIRFIAAAAQAAAAEAgGAAAAH/P/"
    #     "YQAAAB1JREFUOI1j/P///38GCgATJZpHDRg1YNSAwWQAAGvKBByn4XVTAAAAAElFTkSuQmCC"
    # )
    async def get(self, request: Request) -> Response:
        return FileResponse(FAVICON_PATH, media_type="image/x-icon")
        # return Response(Favicon.favicon, media_type="image/x-icon")


configuration.add_route("/favicon.ico", Favicon)



from typing import Optional

from ruteni import configuration
from starlette.responses import Response

COOKIE_SECURE: bool = configuration.get(
    "RUTENI_COOKIE_SECURE",
    cast=bool,
    default=(not configuration.is_devel),
)
COOKIE_HTTPONLY: bool = configuration.get(
    "RUTENI_COOKIE_HTTPONLY", cast=bool, default=True
)
COOKIE_PATH: str = configuration.get("RUTENI_COOKIE_PATH", default="/")
COOKIE_DOMAIN: Optional[str] = configuration.get("RUTENI_COOKIE_DOMAIN", default=None)
COOKIE_SAMESITE: str = configuration.get("RUTENI_COOKIE_SAMESITE", default="lax")


def set_cookie(response: Response, name: str, value: str) -> None:
    response.set_cookie(
        name,
        value,
        path=COOKIE_PATH,
        domain=COOKIE_DOMAIN,
        secure=COOKIE_SECURE,
        httponly=COOKIE_HTTPONLY,
        samesite=COOKIE_SAMESITE,
    )



from __future__ import annotations

from webcolors import CSS3_NAMES_TO_HEX, HEX_COLOR_RE


class Color(str):
    def __new__(cls, p_string: str) -> Color:
        if p_string not in CSS3_NAMES_TO_HEX and not HEX_COLOR_RE.match(p_string):
            raise ValueError(f"invalid color {p_string}")
        return str.__new__(cls, p_string)



from ruteni import configuration
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.endpoints import HTTPEndpoint


class Robots(HTTPEndpoint):
    robots_txt = "User-agent: *\nDisallow:"

    async def get(self, request: Request) -> Response:
        return PlainTextResponse(Robots.robots_txt)


configuration.add_route("/robots.txt", Robots)



import logging

from ruteni.plugins.oauth import oauth

logger = logging.getLogger(__name__)

twitter = oauth.register(
    name="twitter",
    api_base_url="https://api.twitter.com/1.1/",
    request_token_url="https://api.twitter.com/oauth/request_token",
    access_token_url="https://api.twitter.com/oauth/access_token",
    authorize_url="https://api.twitter.com/oauth/authenticate",
)

# TODO: implement…
raise NotImplementedError("FIXME")

logger.info("loaded")



import unittest
from pathlib import Path

from ruteni import ROOTNS
from ruteni.plugins.pwa import Display, ProgressiveWebApp, RelatedApplication
from ruteni.utils.color import Color
from ruteni.utils.icon import PngIcon
from urlobject.path import URLPath


class ProgressiveWebAppTestCase(unittest.TestCase):
    def test_get_manifest(self) -> None:
        icons = [
            PngIcon(
                ROOTNS,
                URLPath(f"{size}x{size}.png"),
                Path(f"tests/icons/icon-{size}x{size}.png"),
            )
            for size in (72, 96, 120, 128, 144, 152, 180, 192, 384, 512)
        ]

        # with self.assertRaises(UnknownColor):
        pwa = ProgressiveWebApp(
            "app_name",
            1,
            display=Display.BROWSER,
            theme_color=Color("white"),
            background_color=Color("#ffffff"),
        )
        pwa.add_i18n(
            "en-US",
            full_name="full name",
            short_name="short name",
            description="description",
        )
        pwa.add_i18n(
            "fr-FR",
            full_name="nom entier",
            short_name="nom court",
            description="description",
        )

        for icon in icons:
            pwa.add_icon(icon)

        pwa.add_related_application(
            RelatedApplication(platform="play", id="com.google.samples.apps.iosched")
        )

        with self.assertRaises(KeyError):
            pwa.get_manifest("de-DE")

        self.assertEqual(
            pwa.get_manifest("en-US"),
            {
                "lang": "en-US",
                "icons": [
                    {"src": "/72x72.png", "sizes": "72x72", "type": "image/png"},
                    {"src": "/96x96.png", "sizes": "96x96", "type": "image/png"},
                    {"src": "/120x120.png", "sizes": "120x120", "type": "image/png"},
                    {"src": "/128x128.png", "sizes": "128x128", "type": "image/png"},
                    {"src": "/144x144.png", "sizes": "144x144", "type": "image/png"},
                    {"src": "/152x152.png", "sizes": "152x152", "type": "image/png"},
                    {"src": "/180x180.png", "sizes": "180x180", "type": "image/png"},
                    {"src": "/192x192.png", "sizes": "192x192", "type": "image/png"},
                    {"src": "/384x384.png", "sizes": "384x384", "type": "image/png"},
                    {"src": "/512x512.png", "sizes": "512x512", "type": "image/png"},
                ],
                "start_url": "/app/app_name/",
                "name": "full name",
                "short_name": "short name",
                "description": "description",
                "display": "browser",
                "theme_color": "white",
                "background_color": "#ffffff",
                "related_applications": [
                    {"platform": "play", "id": "com.google.samples.apps.iosched"}
                ],
            },
        )

        self.assertEqual(
            pwa.get_manifest("fr-FR"),
            {
                "lang": "fr-FR",
                "icons": [
                    {"src": "/72x72.png", "sizes": "72x72", "type": "image/png"},
                    {"src": "/96x96.png", "sizes": "96x96", "type": "image/png"},
                    {"src": "/120x120.png", "sizes": "120x120", "type": "image/png"},
                    {"src": "/128x128.png", "sizes": "128x128", "type": "image/png"},
                    {"src": "/144x144.png", "sizes": "144x144", "type": "image/png"},
                    {"src": "/152x152.png", "sizes": "152x152", "type": "image/png"},
                    {"src": "/180x180.png", "sizes": "180x180", "type": "image/png"},
                    {"src": "/192x192.png", "sizes": "192x192", "type": "image/png"},
                    {"src": "/384x384.png", "sizes": "384x384", "type": "image/png"},
                    {"src": "/512x512.png", "sizes": "512x512", "type": "image/png"},
                ],
                "start_url": "/app/app_name/",
                "name": "nom entier",
                "short_name": "nom court",
                "description": "description",
                "display": "browser",
                "theme_color": "white",
                "background_color": "#ffffff",
                "related_applications": [
                    {"platform": "play", "id": "com.google.samples.apps.iosched"}
                ],
            },
        )



from typing import List, Optional

from starlette.middleware.base import RequestResponseEndpoint
from starlette.routing import Route
from starlette.staticfiles import PathLike
from urlobject.path import URLPath

from . import EndpointTransform, configuration


class URLNamespace:
    def __init__(
        self,
        type: str,
        name: str,
        version: int,
        ns: URLPath,
        *,
        transform: Optional[EndpointTransform] = None,
        include_in_schema: bool = True,
    ) -> None:
        """
        from starlette.authentication import requires
        transform = requires(scopes, status_code, redirect)
        """
        self.type = type
        self.name = name
        self.version = version
        self.ns = ns
        self.transform = transform
        self.include_in_schema = include_in_schema
        self.urls: dict[str, str] = {}

    def add_route(
        self,
        path: URLPath,
        endpoint: RequestResponseEndpoint,
        *,
        methods: List[str] = None,
        name: Optional[str] = None,
    ) -> Route:
        real_endpoint = self.transform(endpoint) if self.transform else endpoint
        real_path = self.ns.add(path)
        if name:
            self.urls[name] = real_path
        configuration.add_route(
            real_path,
            real_endpoint,
            methods=methods,
            name=f"{self.name}-{self.version}-{self.type}-{name}" if name else None,
            include_in_schema=self.include_in_schema,
        )

    def add_file_route(
        self,
        path: URLPath,
        filename: PathLike,
        *,
        media_type: str = None,
        name: str = None,
    ) -> Route:
        return configuration.add_file_route(
            path=path, filename=filename, ns=self.ns, media_type=media_type, name=name
        )

    def url_for(self, name: str) -> Optional[str]:
        return self.urls.get(name, None)



import logging

from ruteni.plugins.database import metadata
from sqlalchemy import (
    Column,
    DateTime,
    Index,
    Integer,
    String,
    Table,
    event,
    func,
    text,
)
from sqlalchemy.engine.base import Connection

logger = logging.getLogger(__name__)

locales = Table(
    "locales",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("code", String(28), nullable=False),
    Column("added_at", DateTime, nullable=False, server_default=func.now()),
    Column("disabled_at", DateTime, default=None),
)

Index(
    "ix_locales_code_not_disabled",
    locales.c.code,
    unique=True,
    sqlite_where=locales.c.disabled_at.is_(None),
    postgresql_where=locales.c.disabled_at.is_(None),
)


def after_create(target: Table, connection: Connection, **kwargs):  # type: ignore
    connection.execute(
        text("INSERT INTO %s (code) VALUES ('en-US'), ('fr-FR')" % target.name)
    )


event.listen(locales, "after_create", after_create)

# import pycountry
# for lang in pycountry.languages:
#     if hasattr(lang, 'iso639_1_code'):
#         print(lang.iso639_1_code)
#     print(lang)


logger.info("loaded")



import argparse
import tempfile
from pathlib import Path

import uvicorn

from ruteni import Ruteni, configuration
from ruteni.utils.jwkset import KeyCollection

configuration.set("RUTENI_ENV", "development")

parser = argparse.ArgumentParser()
parser.add_argument(
    "-d",
    "--test-dir",
    type=lambda p: Path(p).absolute(),
    default=Path(tempfile.mkdtemp()).absolute(),
    help="Path to the test directory",
)
parser.add_argument(
    "-s",
    "--static-dir",
    required=False,
    help="Path to the static directory",
)
args = parser.parse_args()

PRIVATE_KEYS = str(args.test_dir / "private_keys.json")
PUBLIC_KEYS = str(args.test_dir / "public_keys.json")
DB_URL = "sqlite:///" + str(args.test_dir / "test.db")

key_collection = KeyCollection(args.test_dir, True)
key_collection.generate()
key_collection.export(PUBLIC_KEYS, PRIVATE_KEYS)

if args.static_dir:
    configuration.set_static_dir(args.static_dir)
else:
    configuration.set_static_dir(Path(__file__).parent / "static")


configuration.set("RUTENI_DATABASE_URL", DB_URL)
configuration.set("RUTENI_SESSION_SECRET_KEY", "secret-key")
configuration.set("RUTENI_JWKEYS_FILE", PRIVATE_KEYS)
configuration.set("RUTENI_VERIFICATION_ABUSE_URL", "<abuse_url>")
configuration.set("RUTENI_VERIFICATION_FROM_ADDRESS", "Accot <contact@accot.fr>")
configuration.set("RUTENI_SITE_NAME", "Ruteni test")

import ruteni.apps.store
import ruteni.plugins.auth.session
import ruteni.plugins.blaze
import ruteni.plugins.logging
import ruteni.plugins.registration
import ruteni.plugins.robots
import ruteni.plugins.security
import ruteni.plugins.socketio

print(DB_URL, sorted(configuration.get_route_paths()))

app = Ruteni()
uvicorn.run(app, host="127.0.0.1", port=8000)



from ruteni.plugins.database import metadata
from sqlalchemy import Column, DateTime, Integer, Table, event, func, text
from sqlalchemy.engine.base import Connection

tos = Table(
    "tos",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("date", DateTime, nullable=False, server_default=func.now()),
)


def after_create(target: Table, connection: Connection, **kwargs):  # type: ignore
    connection.execute(text("INSERT INTO %s DEFAULT VALUES" % target.name))


event.listen(tos, "after_create", after_create)


# from ruteni.plugins.tos import tos
# tos = Integer(required=True, validate=validate.Range(min=0))
# current_tos = await database.fetch_val(select(func.max(tos.c.id)))
# if current_tos != signin["tos"]:
#    raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)



import enum
import logging

import pyrfc3339
from ruteni import configuration
from ruteni.api import WebApi
from ruteni.plugins.database import database, metadata
from ruteni.plugins.users import users
from sqlalchemy import Column, DateTime, Enum, ForeignKey, Integer, String, Table
from starlette.requests import Request
from starlette.responses import Response

logger = logging.getLogger(__name__)


class Level(enum.Enum):
    trace = 1
    debug = 2
    info = 3
    warn = 4
    error = 5


logs = Table(
    "logs",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey(users.c.id)),
    Column("level", Enum(Level), nullable=False),
    Column("logger", String(255), nullable=False),
    Column("message", String, nullable=False),
    Column("timestamp", DateTime, nullable=False),
    Column("stacktrace", String),
)


async def log(request: Request) -> Response:
    """
    {
        "logs": [
            {
                "message": "name-update2",
                "level": <Level.warn: 4>,
                "logger": "registration-homepage",
                "timestamp": datetime.datetime(
                    2021, 10, 1, 21, 27, 55, 700000, tzinfo=datetime.timezone.utc
                ),
                "stacktrace": "    at HTMLInputElement.name-update (http://127.0.0.1:8000/static/components/registration-homepage/index.js:25:28)",
                "user_id": None,
            }
        ]
    }
    """
    body = await request.json()

    for log in body["logs"]:
        log["level"] = Level[log["level"]]
        log["timestamp"] = pyrfc3339.parse(log["timestamp"])
        if log["stacktrace"] == "":
            log["stacktrace"] = None
        log["user_id"] = request.user.id if request.user.is_authenticated else None

    await database.execute_many(logs.insert(), body["logs"])
    return Response()


api = WebApi("logging", 1)
api.add_route("log", log, methods=["POST"], name="log")

configuration.add_static_resource_mount("logging", __name__)

logger.info("loaded")



import logging

from aiosmtpd.controller import Controller
from ruteni import configuration
from starlette.applications import Starlette

logger = logging.getLogger(__name__)

# TODO: implement…
raise NotImplementedError("FIXME")


class SMTPd:
    def __init__(self) -> None:
        self.controller = Controller()

    async def startup(self, starlette: Starlette) -> None:
        logger.info("started")

    async def shutdown(self, starlette: Starlette) -> None:
        logger.info("stopped")


smtpd = SMTPd()
configuration.add_service("smtpd", smtpd.startup, smtpd.shutdown)

logger.info("loaded")



import unittest

from starlette.testclient import TestClient

from .config import (
    USER_EMAIL,
    USER_LOCALE,
    USER_NAME,
    clear_database,
    test_env,
)

with test_env:
    from ruteni import Ruteni
    from ruteni.plugins.groups import (
        GroupStatus,
        InvalidExpressionException,
        UnknownGroupException,
        add_group,
        add_user_to_group,
        test_group_condition,
    )
    from ruteni.plugins.users import UnknownUserException, add_user


class GroupsTestCase(unittest.IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.ruteni = Ruteni()

    def tearDown(self) -> None:
        clear_database()

    async def test_add_user_to_group(self) -> None:
        with TestClient(self.ruteni):
            user_info = await add_user(USER_NAME, USER_EMAIL, USER_LOCALE)
            user_id = user_info.id
            with self.assertRaises(UnknownGroupException):
                await add_user_to_group(user_id, "foo")
            await add_group("foo")
            self.assertEqual(await add_user_to_group(user_id, "foo"), 1)

    async def test_test_group_condition(self) -> None:
        with TestClient(self.ruteni):

            with self.assertRaises(UnknownUserException):
                await test_group_condition(2, "foo")

            user_info = await add_user(USER_NAME, USER_EMAIL, USER_LOCALE)
            user_id = user_info.id

            self.assertFalse(await test_group_condition(user_id, "FALSE"))
            self.assertTrue(await test_group_condition(user_id, "TRUE"))

            with self.assertRaises(InvalidExpressionException):
                await test_group_condition(user_id, "(")

            self.assertFalse(await test_group_condition(user_id, "qux"))
            self.assertTrue(await test_group_condition(user_id, "not qux"))

            await add_group("foo")
            await add_group("bar")
            await add_group("baz")

            await add_user_to_group(user_id, "foo")
            self.assertTrue(await test_group_condition(user_id, "foo"))
            self.assertTrue(await test_group_condition(user_id, "foo:member"))
            self.assertFalse(await test_group_condition(user_id, "not foo:member"))
            self.assertFalse(await test_group_condition(user_id, "foo:manager"))
            self.assertTrue(await test_group_condition(user_id, "not foo:manager"))
            self.assertFalse(await test_group_condition(user_id, "foo:owner"))
            self.assertTrue(await test_group_condition(user_id, "not foo:owner"))
            self.assertFalse(await test_group_condition(user_id, "foo and bar"))
            self.assertTrue(await test_group_condition(user_id, "foo or bar"))
            self.assertTrue(await test_group_condition(user_id, "foo and not bar"))
            self.assertFalse(await test_group_condition(user_id, "foo and qux"))
            self.assertTrue(await test_group_condition(user_id, "foo or qux"))
            self.assertTrue(await test_group_condition(user_id, "foo and not qux"))
            self.assertFalse(await test_group_condition(user_id, "foo and bar and baz"))
            self.assertFalse(await test_group_condition(user_id, "foo and bar and qux"))

            await add_user_to_group(user_id, "bar", GroupStatus.manager)
            self.assertTrue(await test_group_condition(user_id, "foo"))
            self.assertTrue(await test_group_condition(user_id, "foo and bar"))
            self.assertTrue(await test_group_condition(user_id, "foo and bar:manager"))
            self.assertFalse(await test_group_condition(user_id, "foo and bar:owner"))
            self.assertFalse(await test_group_condition(user_id, "foo and bar and baz"))
            self.assertTrue(
                await test_group_condition(user_id, "foo and bar and not baz")
            )
            self.assertTrue(
                await test_group_condition(user_id, "foo and (qux or not baz)")
            )

            await add_user_to_group(user_id, "baz", GroupStatus.owner)
            self.assertTrue(await test_group_condition(user_id, "foo"))
            self.assertTrue(await test_group_condition(user_id, "foo and bar"))
            self.assertTrue(await test_group_condition(user_id, "foo and bar and baz"))
            self.assertTrue(await test_group_condition(user_id, "baz:manager"))
            self.assertTrue(await test_group_condition(user_id, "baz:owner"))


if __name__ == "__main__":
    unittest.main()



import logging
from typing import Callable, Tuple

from pkg_resources import resource_filename
from ruteni import ROOTNS, configuration
from ruteni.plugins.security import ContentSecurityPolicy, add_html_file_route
from ruteni.plugins.session import UserType, get_session_user
from starlette.applications import Starlette
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    BaseUser,
    UnauthenticatedUser,
)
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.requests import Request

logger = logging.getLogger(__name__)

AuthPair = Tuple[AuthCredentials, BaseUser]
AuthFunc = Callable[[UserType], AuthPair]

providers: dict[str, AuthFunc] = {}


def register_identity_provider(name: str, func: AuthFunc) -> None:
    providers[name] = func


class SessionAuthenticationBackend(AuthenticationBackend):
    async def authenticate(self, request: Request) -> AuthPair:
        user = get_session_user(request)
        if user:
            for provider, func in providers.items():
                if user["provider"] == provider:
                    return func(user)
            logger.warn(f'unknown identity provider: {user["provider"]}')
        return AuthCredentials(), UnauthenticatedUser()


async def startup(starlette: Starlette) -> None:
    if len(providers) == 0:
        logger.warn("no identity provider was added with register_identity_provider")
    logger.info("started")


async def shutdown(starlette: Starlette) -> None:
    logger.info("stopped")


content_security_policy = ContentSecurityPolicy(
    script=True, connect=True, style=True, img=True
)

add_html_file_route(
    "ap/signin",
    resource_filename(__name__, "/resources/index.html"),
    content_security_policy,
    ns=ROOTNS,
)

configuration.add_static_resource_mount("auth", __name__)

configuration.add_service("auth", startup, shutdown)

configuration.add_middleware(
    AuthenticationMiddleware, backend=SessionAuthenticationBackend()
)

logger.info("loaded")



import setuptools

# with open("README.md", "r") as fh:
#    long_description = fh.read()

setuptools.setup(
    name="ruteni",
    version="0.5.1",
    author="Johnny Accot",
    description="Thin layer over Starlette",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    py_modules=["ruteni"],
    package_data={
        "": [
            "resources/*",
            "templates/*",
            "static/*",
            "static/*/*",
            "static/*/*/*",
            "static/*/*/*/*",
        ]
    },
    install_requires=[
        "aiodns",
        "aioredis",
        "aiosmtplib",
        "aiosmtpd",
        "aiosqlite",
        "anyio",
        "apscheduler",
        "argon2_cffi",
        "authlib",
        "babel",
        "boolean.py",
        "databases",
        "html5lib",
        "httpx",
        "itsdangerous",
        "jinja2",
        "jwcrypto",
        "limits",
        "marshmallow",
        "python-multipart",
        "paramiko",
        "pillow",
        "python-socketio",
        "pyrfc3339",
        "sqlalchemy",
        "sqlalchemy-utils",
        "starlette",
        "tabulate",
        "transitions",
        "urlobject",
        "uvicorn",
        "webcolors",
        "websockets",
        "werkzeug",
        "zxcvbn",
    ],
    test_suite="tests.build_test_suite",
)



import enum
import logging
from typing import NamedTuple, Sequence

from boolean import BooleanAlgebra, boolean
from ruteni.plugins.database import database, metadata
from ruteni.plugins.users import assert_user_exists, users
from sqlalchemy import (
    Column,
    DateTime,
    Enum,
    ForeignKey,
    Index,
    Integer,
    String,
    Table,
    and_,
    event,
    func,
    text,
)
from sqlalchemy.engine.base import Connection
from sqlalchemy.sql import select

logger = logging.getLogger(__name__)


class UnknownGroupException(Exception):
    def __init__(self, group_name: str) -> None:
        super().__init__(f"unknown group: {group_name}")
        self.group_name = group_name


class InvalidExpressionException(Exception):
    def __init__(self, expression: str) -> None:
        super().__init__(f"unknown expression: {expression}")
        self.expression = expression


class GroupStatus(enum.Enum):
    member = 1
    manager = 2
    owner = 3


groups = Table(
    "groups",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("name", String(28), nullable=False),
    Column("added_at", DateTime, nullable=False, server_default=func.now()),
    Column("disabled_at", DateTime, default=None),
)

Index(
    "ix_groups_name_not_disabled",
    groups.c.name,
    unique=True,
    sqlite_where=groups.c.disabled_at.is_(None),
    postgresql_where=groups.c.disabled_at.is_(None),
)

memberships = Table(
    "memberships",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey(users.c.id), nullable=False),
    Column("group_id", Integer, ForeignKey(groups.c.id), nullable=False),
    Column("status", Enum(GroupStatus), nullable=False),
    # Column("added_by", Integer, ForeignKey(users.c.id), nullable=False),
    Column("added_at", DateTime, nullable=False, server_default=func.now()),
    # Column("disabled_by", Integer, ForeignKey(users.c.id), default=None),
    Column("disabled_at", DateTime, default=None),
)

Index(
    "ix_memberships_user_group_not_disabled",
    memberships.c.user_id,
    memberships.c.group_id,
    unique=True,
    sqlite_where=memberships.c.disabled_at.is_(None),
    postgresql_where=memberships.c.disabled_at.is_(None),
)


def after_create(target: Table, connection: Connection, **kwargs):  # type: ignore
    connection.execute(text("INSERT INTO %s (name) VALUES ('admin')" % target.name))


event.listen(groups, "after_create", after_create)


class MembershipInfo(NamedTuple):
    group: str
    status: GroupStatus


async def get_user_groups(user_id: int) -> Sequence[str]:
    await assert_user_exists(user_id)
    return tuple(
        membership_info.group for membership_info in await get_user_memberships(user_id)
    )


async def get_user_memberships(user_id: int) -> Sequence[MembershipInfo]:
    await assert_user_exists(user_id)
    return tuple(
        MembershipInfo(row["name"], row["status"])
        for row in await database.fetch_all(
            select([groups.c.name, memberships.c.status])
            .select_from(memberships.join(groups).join(users))
            .where(
                and_(
                    memberships.c.user_id == user_id,
                    users.c.disabled_at.is_(None),
                    groups.c.disabled_at.is_(None),
                    memberships.c.disabled_at.is_(None),
                )
            )
        )
    )


async def add_group(name: str) -> int:
    return await database.execute(groups.insert().values(name=name))


async def add_user_to_group(
    user_id: int, group_name: str, status: GroupStatus = GroupStatus.member
) -> int:
    await assert_user_exists(user_id)

    # TODO: try to use insert(memberships).from_select(...)
    group_id = await database.fetch_val(
        select([groups.c.id]).where(
            and_(groups.c.name == group_name, groups.c.disabled_at.is_(None))
        )
    )
    if group_id is None:
        raise UnknownGroupException(group_name)

    membership_id = await database.execute(
        memberships.insert().values(user_id=user_id, group_id=group_id, status=status)
    )
    return membership_id


def eval_boolean_expression(
    expression: boolean.Expression, predicates: set[str]
) -> bool:
    if isinstance(expression, boolean.Symbol):
        logger.debug("SYM", expression.obj in predicates)
        return expression.obj in predicates
    elif isinstance(expression, boolean.AND):
        logger.debug("AND", expression.args)
        return all(eval_boolean_expression(arg, predicates) for arg in expression.args)
    elif isinstance(expression, boolean.OR):
        logger.debug("OR", expression.args)
        return any(eval_boolean_expression(arg, predicates) for arg in expression.args)
    elif isinstance(expression, boolean.NOT):
        logger.debug("NOT", expression.args)
        return not eval_boolean_expression(expression.args[0], predicates)
    elif isinstance(expression, boolean._TRUE):
        logger.debug("TRUE")
        return True
    elif isinstance(expression, boolean._FALSE):
        logger.debug("FALSE")
        return False
    else:
        raise Exception("unexpected error")


async def test_group_condition(user_id: int, condition: str) -> bool:
    await assert_user_exists(user_id)

    algebra = BooleanAlgebra()
    try:
        expression = algebra.parse(condition, simplify=False)
    except boolean.ParseError:
        raise InvalidExpressionException(condition)

    for symbol in expression.symbols:
        # TODO: test correct symbol format (zero or one colon)
        if ":" not in symbol.obj:
            symbol.obj += ":member"

    user_memberships = set(
        f"{membership_info.group}:{status.name}"
        for membership_info in await get_user_memberships(user_id)
        for status in GroupStatus
        if status.value <= membership_info.status.value
    )

    logger.debug(expression, user_memberships)
    return eval_boolean_expression(expression, user_memberships)


logger.info("loaded")



from __future__ import annotations

from typing import Optional

from urlobject.path import URLPath

from . import EndpointTransform, ROOTNS
from .urlns import URLNamespace

APINS = ROOTNS.add("api")


class WebApi(URLNamespace):
    """
    Manage a web API
    """

    url_format = "/api/{name}/v{version}"

    def __init__(
        self,
        name: str,
        version: int,
        *,
        transform: Optional[EndpointTransform] = None,
    ) -> None:
        ns = URLPath(WebApi.url_format.format(name=name, version=version))
        super().__init__("api", name, version, ns, transform=transform)



import logging
from pathlib import Path

import html5lib
from html5lib.treewalkers import getTreeWalker

from ruteni import STATICNS, configuration

logger = logging.getLogger(__name__)

COMPONENTSNS = STATICNS.add("components")

dependencies: dict[str, set[str]] = {}


def load_component_dir(directory: Path) -> None:
    for node in directory.iterdir():
        if node.is_file():
            if node.suffix == ".js":
                configuration.add_file_route(node.name + ".js", node, ns=COMPONENTSNS)
            else:
                logger.warn(f"file {node} ignored")
        else:
            index_js = node.joinpath("index.js")
            if index_js.exists():
                configuration.add_file_route(
                    node.name + ".js", index_js, ns=COMPONENTSNS
                )

                index_html = node.joinpath("index.html")
                if index_html.exists():
                    with open(index_html, "rb") as f:
                        fragment = html5lib.parseFragment(f)
                    walker = getTreeWalker("etree")
                    stream = walker(fragment)
                    dependencies[node.name] = set()
                    for element in stream:
                        if element["type"] == "StartTag":
                            if "-" in element["name"]:
                                dependencies[node.name].add(element["name"])
                    configuration.add_file_route(
                        node.name + ".html", index_html, ns=COMPONENTSNS
                    )

                index_css = node.joinpath("index.css")
                if index_css.exists():
                    configuration.add_file_route(
                        node.name + ".css", index_css, ns=COMPONENTSNS
                    )
            else:
                logger.warn(f"missing index.js in '{node}'")


logger.info("loaded")



import logging
from typing import Sequence

from babel.core import Locale, get_locale_identifier
from starlette.requests import Request

logger = logging.getLogger(__name__)


def parse_accept_language(accept_language: str) -> Sequence[Locale]:
    """
    TODO: properly manage format https://pypi.org/project/parse-accept-language/
    """
    locales: list[Locale] = []
    for locale in accept_language.split(","):
        if ";" in locale:
            locale, quality = locale.split(";")
        locales.append(Locale.parse(locale, sep="-"))
    return locales


def get_locale_from_request(request: Request, available: Sequence[Locale]) -> Locale:
    accept_language = request.headers.get("Accept-Language", None)
    if accept_language is None:
        return Locale("fr", "FR")  # TODO: pref

    preferred = parse_accept_language(accept_language)
    locale = Locale.negotiate(
        [str(locale) for locale in preferred], [str(locale) for locale in available]
    )
    logger.debug(f"negotiate_locale {preferred} from {available} -> {locale}")
    return locale


def get_html_lang(locale: Locale) -> str:
    return get_locale_identifier(
        (locale.language, locale.territory, locale.script, locale.variant), sep="-"
    )



import logging
from datetime import datetime, timedelta, timezone
from secrets import token_hex
from typing import Any, Mapping, NamedTuple, Optional

from ruteni import configuration
from ruteni.plugins.database import database, metadata
from ruteni.plugins.groups import get_user_groups
from ruteni.plugins.jwkeys import jwkeys
from ruteni.plugins.users import UserInfo, get_user_by_id, users
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Table, and_, func
from sqlalchemy.sql import select

logger = logging.getLogger(__name__)

REFRESH_TOKEN_LENGTH: int = configuration.get(
    "RUTENI_REFRESH_TOKEN_LENGTH", cast=int, default=32
)
MAX_CLIENT_ID_LENGTH: int = configuration.get(
    "RUTENI_MAX_CLIENT_ID_LENGTH", cast=int, default=255
)
REFRESH_TOKEN_EXPIRATION: int = configuration.get(
    "RUTENI_REFRESH_TOKEN_EXPIRATION", cast=int, default=14 * 24 * 60 * 60  # 14 days
)
ACCESS_TOKEN_EXPIRATION: int = configuration.get(
    "RUTENI_ACCESS_TOKEN_EXPIRATION", cast=int, default=60 * 60  # 1 hour
)

refresh_tokens = Table(
    "refresh_tokens",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("user_id", Integer, ForeignKey(users.c.id), nullable=False),
    Column("token", String(2 * REFRESH_TOKEN_LENGTH), unique=True, nullable=False),
    Column("client_id", String(MAX_CLIENT_ID_LENGTH), nullable=False),
    Column("issued_at", DateTime, nullable=False),
    Column("expires_at", DateTime, nullable=False),
    Column("revoked_at", DateTime, default=None),
)

access_tokens = Table(
    "access_tokens",
    metadata,
    Column("id", Integer, primary_key=True),
    Column(
        "refresh_token_id", Integer, ForeignKey(refresh_tokens.c.id), nullable=False
    ),
    Column("issued_at", DateTime, nullable=False),
    Column("expires_at", DateTime, nullable=False),
)


def now() -> datetime:
    "Have only one way to reference now. Also useful for unittest mock."
    return datetime.now(timezone.utc).astimezone()


def new_refresh_token() -> str:
    # TODO: though very unlikely, check that the generated token is not in the table
    return token_hex(REFRESH_TOKEN_LENGTH)


class AccessToken(NamedTuple):
    token: str
    claims: dict
    expires_at: datetime


async def create_access_claims(
    refresh_token_id: int,
    refresh_token_expires_at: datetime,
    access_token_id: int,
    issuer: str,
    issued_at: datetime,
    expires_at: datetime,
    user_id: int,
    client_id: str,
) -> tuple[UserInfo, dict]:
    user_info = await get_user_by_id(user_id)
    assert user_info
    scopes = await get_user_groups(user_id)
    return user_info, dict(
        sub=str(user_id),
        jti=str(access_token_id),
        iat=int(issued_at.timestamp()),
        exp=int(expires_at.timestamp()),
        rte=int(refresh_token_expires_at.timestamp()),
        orig_iat=0,  # TODO: https://stackoverflow.com/questions/50258431/
        iss=issuer,
        email=user_info.email,
        display_name=user_info.display_name,
        scope=scopes,
        client_id=client_id,
    )


async def create_access_token(
    refresh_token_id: int,
    refresh_token_expires_at: datetime,
    user_id: int,
    issuer: str,
    client_id: str,
) -> tuple[UserInfo, AccessToken]:
    issued_at = now()
    expires_at = issued_at + timedelta(seconds=ACCESS_TOKEN_EXPIRATION)
    access_token_id = await database.execute(
        access_tokens.insert().values(
            refresh_token_id=refresh_token_id,
            issued_at=issued_at,
            expires_at=expires_at,
        )
    )
    user, claims = await create_access_claims(
        refresh_token_id,
        refresh_token_expires_at,
        access_token_id,
        issuer,
        issued_at,
        expires_at,
        user_id,
        client_id,
    )
    token = jwkeys.create_token(claims)

    return user, AccessToken(token, claims, expires_at)


async def create_refresh_token(
    user_id: int, issuer: str, client_id: str
) -> tuple[UserInfo, AccessToken, str]:
    issued_at = now()
    expires_at = issued_at + timedelta(seconds=REFRESH_TOKEN_EXPIRATION)
    refresh_token = new_refresh_token()
    refresh_token_id = await database.execute(
        refresh_tokens.insert().values(
            token=refresh_token,
            issued_at=issued_at,
            expires_at=expires_at,
            user_id=user_id,
            client_id=client_id,
        )
    )
    user, access_token = await create_access_token(
        refresh_token_id, expires_at, user_id, issuer, client_id
    )

    return user, access_token, refresh_token


async def revoke_refresh_token(user_id: int, refresh_token: str) -> bool:
    rowcount = await database.execute(
        refresh_tokens.update()
        .where(
            and_(
                refresh_tokens.c.token == refresh_token,
                refresh_tokens.c.user_id == user_id,
                refresh_tokens.c.revoked_at.is_(None),
            )
        )
        .values(revoked_at=func.now())
    )
    return rowcount > 0


async def select_refresh_token(refresh_token: str) -> Optional[Mapping[Any, Any]]:
    return await database.fetch_one(
        select([refresh_tokens])
        .select_from(users.join(refresh_tokens))
        .where(refresh_tokens.c.token == refresh_token)
    )


def get_user_id_from_access_token(access_token: str, issuer: str) -> int:
    claims = jwkeys.get_claims(access_token, issuer)
    return int(claims["sub"])


logger.info("loaded")



import json
import logging
from base64 import b64decode
from typing import Any, Optional

import itsdangerous
from itsdangerous.exc import BadTimeSignature, SignatureExpired
from ruteni import configuration
from starlette.middleware.sessions import SessionMiddleware
from starlette.requests import Request, cookie_parser

logger = logging.getLogger(__name__)

USER_SESSION_NAME: str = configuration.get("RUTENI_USER_SESSION_NAME", default="user")
SECRET_KEY = configuration.get("RUTENI_SESSION_SECRET_KEY")
MAX_AGE = configuration.get(
    "RUTENI_SESSION_MAX_AGE", cast=int, default=2 * 7 * 24 * 60 * 60  # 2 weeks
)
COOKIE_NAME = configuration.get("RUTENI_SESSION_COOKIE_NAME", default="session")
SAME_SITE = configuration.get("RUTENI_SESSION_SAME_SITE", default="lax")

session_middleware = configuration.add_middleware(
    SessionMiddleware,
    https_only=not configuration.is_devel,
    max_age=MAX_AGE,
    same_site=SAME_SITE,
    secret_key=SECRET_KEY,
    session_cookie=COOKIE_NAME,
)

UserType = dict[str, Any]

signer = itsdangerous.TimestampSigner(SECRET_KEY)


def set_session_user(request: Request, user: UserType) -> None:
    request.session[USER_SESSION_NAME] = user


def del_session_user(request: Request) -> None:
    request.session.pop(USER_SESSION_NAME, None)


def get_session_user(request: Request) -> Optional[UserType]:
    return request.session.get(USER_SESSION_NAME, None)


def verify_and_decode_cookie(cookie: str) -> Optional[dict]:
    try:
        data = signer.unsign(cookie, max_age=MAX_AGE)
    except (BadTimeSignature, SignatureExpired):
        return None
    else:
        return json.loads(b64decode(data))


def get_user_from_cookie(cookie: str) -> Optional[UserType]:
    session = verify_and_decode_cookie(cookie)
    return (
        session[USER_SESSION_NAME]
        if session is not None and USER_SESSION_NAME in session
        else None
    )


def get_user_from_environ(environ: dict[str, str]) -> Optional[UserType]:
    if "HTTP_COOKIE" in environ:
        cookies = cookie_parser(environ["HTTP_COOKIE"])
        if COOKIE_NAME in cookies:
            return get_user_from_cookie(cookies[COOKIE_NAME])

    # in all other cases, no user was found
    return None


logger.info("loaded")



import logging

from ruteni.plugins.database import database, metadata
from ruteni.plugins.session import get_user_from_environ
from ruteni.plugins.socketio import sio
from ruteni.plugins.users import users
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Table, and_, func
from sqlalchemy_utils import IPAddressType

logger = logging.getLogger(__name__)

connections = Table(
    "connections",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("sid", String(28), nullable=False),
    Column("ip_address", IPAddressType, nullable=False),
    Column("user_id", Integer, ForeignKey(users.c.id), nullable=False),
    Column("opened_at", DateTime, nullable=False, server_default=func.now()),
    Column("closed_at", DateTime, default=None),
)


async def on_connect(sid: str, environ: dict[str, str]) -> bool:
    # async with sio.eio.session(sid) as session:
    #     session["username"] = username

    # get the current user
    user = get_user_from_environ(environ)
    if user is None:
        return False  # reject connection

    await database.execute(
        connections.insert().values(
            sid=sid, user_id=user["id"], ip_address=environ["REMOTE_ADDR"]
        )
    )
    logger.info(f"{user['name']} is connected")
    return True


async def on_disconnect(sid: str) -> None:
    # TODO: could there be identical sids for different connections over time?
    await database.execute(
        connections.update()
        .where(and_(connections.c.sid == sid, connections.c.closed_at.is_(None)))
        .values(closed_at=func.now())
    )
    logger.info(f"{sid}.disconnect")


sio.on("connect", on_connect)
sio.on("disconnect", on_disconnect)

logger.info("loaded")



import logging
import os
import re
from datetime import datetime
from typing import Any

from anyio import open_file
from jwcrypto.common import json_decode
from jwcrypto.jwe import JWE
from jwcrypto.jwk import JWK, InvalidJWKValue, JWKSet
from jwcrypto.jwt import JWT
from tabulate import tabulate

logger = logging.getLogger(__name__)

Claims = dict[str, Any]


class KeySet:
    def __init__(self) -> None:
        self.keyset: JWKSet = None
        self.key: JWK = None

    async def load(self, key_file: str) -> None:
        async with await open_file(key_file) as f:
            content = await f.read()

        self.keyset = JWKSet.from_json(content)

        if len(self.keyset) == 0:
            raise Exception("empty key file")

        for key in self.keyset:
            if self.key is None or self.key.get("kid") < key.get("kid"):
                self.key = key

        logger.debug(f"using key {self.key.get('kid')}")

    def create_token(self, **claims: Claims) -> str:
        header = dict(alg="RS256", typ="JWT", kid=self.key.get("kid"))
        jwt = JWT(header=header, claims=claims)
        jwt.make_signed_token(self.key)
        return jwt.serialize()

    def get_claims(self, token: str) -> Claims:
        jwt = JWT(jwt=token, key=self.keyset)
        return json_decode(jwt.claims)

    def export(self) -> str:
        return self.keyset.export(private_keys=False)


class KeyCollection:
    def __init__(self, key_dir: str, create: bool = False) -> None:

        self.key_dir = key_dir
        self.keys = []

        if not os.path.exists(key_dir):
            if create:
                os.mkdir(self.key_dir)
            else:
                raise Exception(f"key dir {key_dir} does not exist")

        if not os.path.isdir(key_dir):
            logger.error(f"{key_dir} is not a directory")
            return

        for filename in sorted(os.listdir(key_dir)):

            m = re.match(r"(\d+)(?:-(\d+))?", filename)
            if not m:
                logger.error(f"ignoring {filename}")
                continue

            created = datetime.fromtimestamp(int(m.group(1)))
            revoked = datetime.fromtimestamp(int(m.group(2))) if m.group(2) else None

            file_path = os.path.join(key_dir, filename)
            with open(file_path) as f:
                content = f.read()

            try:
                key = JWK.from_json(content)
            except InvalidJWKValue:
                logger.error(f"{file_path} is an invalid key file")
            else:
                try:
                    kid = int(key.get("kid"))
                except ValueError:
                    logger.error(f"{file_path} has an invalid key id")
                else:
                    self.keys.append([filename, key, created, revoked])
                    if revoked:
                        logger.debug(f"key {kid} revoked")

    def generate(self, kty: str = "RSA", size: int = 2048) -> None:
        current_kid = -1
        for filename, key, created, revoked in self.keys:
            kid = int(key.get("kid"))
            if current_kid < kid:
                current_kid = kid
        logger.debug(f"current kid: {current_kid}")

        key = JWK.generate(kty=kty, size=size, kid=str(current_kid + 1))
        content = key.export()

        created = int(datetime.now().timestamp())
        filename = f"{created}.json"
        file_path = os.path.join(self.key_dir, filename)

        with open(file_path, "w") as f:
            f.write(content)

        logger.debug(f"new key file: {file_path}")
        self.keys.append([filename, key, created, None])

    def revoke(self, kid: int) -> None:
        for filename, key, created, revoked in self.keys:
            if kid == key.get("kid"):
                if not revoked:
                    revoked = int(datetime.now().timestamp())
                    print("mv", filename, filename[:-5] + "-" + str(revoked) + ".json")
                else:
                    logger.error(f"{kid} already revoked")
                break

    def as_jwkset(self) -> JWKSet:
        jwkset = JWKSet()
        for filename, key, created, revoked in self.keys:
            if not revoked:
                jwkset.add(key)
        return jwkset

    def export(self, public_json: str, private_json: str) -> None:
        jwkset = self.as_jwkset()

        private_dir = os.path.dirname(private_json)
        if private_dir != "":
            os.makedirs(private_dir, exist_ok=True)

        with open(private_json, "w") as f:
            f.write(jwkset.export(private_keys=True))

        public_dir = os.path.dirname(public_json)
        if public_dir != "":
            os.makedirs(public_dir, exist_ok=True)

        with open(public_json, "w") as f:
            f.write(jwkset.export(private_keys=False))

    def print(self) -> None:
        print(
            tabulate(
                tuple(
                    (filename, key.get("kid"), created, revoked)
                    for filename, key, created, revoked in self.keys
                ),
                headers=("filename", "kid", "created", "revoked"),
            )
        )

    def encrypt(self, payload: str, kid: int) -> str:
        if kid:
            for filename, key, created, revoked in self.keys:
                if int(key.get("kid")) == kid:
                    if revoked:
                        raise Exception(f"key {kid} is revoked")
                    break
            else:
                raise Exception(f"unknown key {kid}")
        else:
            latest_key = None
            for filename, key, created, revoked in self.keys:
                if not revoked and (
                    latest_key is None
                    or int(key.get("kid")) > int(latest_key.get("kid"))
                ):
                    latest_key = key
            if latest_key is None:
                raise Exception("empty keyset")
            key = latest_key

        # logger.debug(f'using key {key.get('kid')}')

        public_key = JWK()
        public_key.import_key(**json_decode(key.export_public()))

        protected_header = {
            "alg": "RSA-OAEP-256",
            "enc": "A256CBC-HS512",
            "typ": "JWE",
            "kid": key.get("kid"),
        }

        jwe = JWE(
            payload.encode("utf-8"),
            recipient=public_key,
            protected=protected_header,
        )
        return jwe.serialize()


if __name__ == "__main__":

    from argparse import ArgumentParser, Namespace

    parser = ArgumentParser(prog="jwkset")
    parser.add_argument("--key-dir", default=".")
    subparsers = parser.add_subparsers(help="sub-command help")

    def do_new(args: Namespace) -> None:
        collection = KeyCollection(args.key_dir, True)
        collection.generate()

    parser_new = subparsers.add_parser("new")
    parser_new.set_defaults(func=do_new)

    def do_list(args: Namespace) -> None:
        collection = KeyCollection(args.key_dir)
        collection.print()

    parser_list = subparsers.add_parser("list")
    parser_list.set_defaults(func=do_list)

    def do_revoke(args: Namespace) -> None:
        collection = KeyCollection(args.key_dir)
        collection.revoke(args.kid)

    parser_revoke = subparsers.add_parser("revoke")
    parser_revoke.add_argument("kid")
    parser_revoke.set_defaults(func=do_revoke)

    def do_export(args: Namespace) -> None:
        collection = KeyCollection(args.key_dir)
        collection.export(args.public_keys, args.private_keys)

    parser_export = subparsers.add_parser("export")
    parser_export.add_argument("--public-keys", default="public-keys.json")
    parser_export.add_argument("--private-keys", default="private-keys.json")
    parser_export.set_defaults(func=do_export)

    def do_encrypt(args: Namespace) -> None:
        collection = KeyCollection(args.key_dir)
        print(collection.encrypt(args.data, args.kid))

    parser_encrypt = subparsers.add_parser("encrypt")
    parser_encrypt.add_argument("-k", "--kid", type=int, default=None)
    parser_encrypt.add_argument("data")
    parser_encrypt.set_defaults(func=do_encrypt)

    args = parser.parse_args()
    args.func(args)



import json
from unittest import IsolatedAsyncioTestCase

from ruteni import Ruteni
from sqlalchemy.sql import select
from starlette import status
from starlette.middleware.cors import ALL_METHODS
from starlette.testclient import TestClient

from .config import clear_database, test_env

with test_env:
    from ruteni.plugins.database import database
    from ruteni.plugins.security import (
        CSP_REPORT_CONTENT_TYPE,
        CSP_REPORT_KEY,
        CSP_REPORT_URI,
        EXPECT_CT_CONTENT_TYPE,
        EXPECT_CT_KEY,
        EXPECT_CT_URI,
        ReportType,
        security_reports,
    )

CSP_REPORT = {
    "csp-report": {
        "document-uri": "http://localhost:8000/app/store/",
        "referrer": "",
        "violated-directive": "script-src-elem",
        "effective-directive": "script-src-elem",
        "original-policy": "default-src 'self'; worker-src 'self'; frame-src 'none'; child-src 'none'; object-src 'none'; require-trusted-types-for 'script'; report-uri /api/logging/v1/csp-report/;",
        "disposition": "enforce",
        "blocked-uri": "inline",
        "line-number": 16,
        "source-file": "http://localhost:8000/app/store/",
        "status-code": 200,
        "script-sample": "",
    }
}
EXPECT_CT_REPORT = {
    "expect-ct-report": {
        "date-time": "2017-05-05T12:45:00Z",
        "hostname": "example.com",
        "port": 443,
        "effective-expiration-date": "2017-05-05T12:45:00Z",
        "foo": "bar",
    }
}


class SecurityTestCase(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.ruteni = Ruteni()

    def tearDown(self) -> None:
        clear_database()

    async def test_csp_report(self) -> None:
        with TestClient(self.ruteni) as client:
            for method in ALL_METHODS:
                if method != "POST":
                    response = client.request(method, CSP_REPORT_URI)
                    self.assertEqual(
                        response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                    )

            response = client.post(
                CSP_REPORT_URI,
                json=CSP_REPORT,
                headers={"Content-Type": CSP_REPORT_CONTENT_TYPE},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.json(), True)
            row = await database.fetch_one(
                select([security_reports]).where(security_reports.c.id == 1)
            )
            assert row  # TODO: self.assert...
            self.assertEqual(row["report"], CSP_REPORT[CSP_REPORT_KEY])
            self.assertEqual(row["type"], ReportType.CSP)

    async def test_expect_ct(self) -> None:
        with TestClient(self.ruteni) as client:
            for method in ALL_METHODS:
                if method != "POST":
                    response = client.request(method, EXPECT_CT_URI)
                    self.assertEqual(
                        response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                    )

            response = client.post(
                EXPECT_CT_URI,
                json=EXPECT_CT_REPORT,
                headers={"Content-Type": EXPECT_CT_CONTENT_TYPE},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            self.assertEqual(response.json(), True)
            row = await database.fetch_one(
                select([security_reports]).where(security_reports.c.id == 1)
            )
            assert row  # TODO: self.assert...
            self.assertEqual(row["report"], EXPECT_CT_REPORT[EXPECT_CT_KEY])
            self.assertEqual(row["type"], ReportType.EXPECT_CT)
